from typing import Optional

SUFFIX_MIME_TYPE_MAP = {
    'ttl': 'text/turtle',
    'trig': 'application/trig',
    'jsonld': 'application/ld+json',
    'xml': 'application/rdf+xml',
    'rdf': 'application/rdf+xml',
    'rdfs': 'application/rdf+xml',
    'owl': 'application/rdf+xml',
    'html': 'text/html',
}


def guess_mime_type(ref: str) -> Optional[str]:
    """
    >>> guess_mime_type('path/to/some.dir/file.jsonld')
    'application/ld+json'
    """
    i: int = ref.rfind('.')
    if i == -1:
        return None
    suffix: str = ref[i + 1:]
    return SUFFIX_MIME_TYPE_MAP.get(suffix)



from typing import Optional, Dict, List, Set, Union, cast
from ..common import warning
from .base import *
from .context import Context, Term, InvalidNestValueError
from .invcontext import get_inverse_context


PRESERVE: str = '@preserve'


class IRIConfusedWithPrefixError(JsonLdError):
    pass


def compact(context: object, doc_data: JsonObject,
        base_iri: Optional[str] = None,
        compact_arrays = True,
        ordered = False) -> JsonObject:
    active_context: Context
    if isinstance(context, Context):
        active_context = context
    else:
        context_url: Optional[str] = context if isinstance(context, str) else None
        if isinstance(context, Dict):
            context = context.get(CONTEXT)
        active_context = Context(base_iri).get_context(cast(object, context), context_url)

    result: JsonObject = compaction(active_context, None, doc_data, compact_arrays, ordered)
    if isinstance(result, List):
        if len(result) == 0:
            result = {}
        else:
            result = {iri_compaction(active_context, GRAPH): result}
    return result


def compaction(
        active_context: Context,
        active_property: Optional[str],
        element: JsonObject,
        compact_arrays = False,
        ordered = False
        ) -> JsonObject:
    # 1)
    type_scoped_context: Context = active_context
    result: Union[Dict, List]
    result_map: JsonMap
    result_list: JsonList

    active_term: Optional[Term] = active_context.terms.get(active_property)
    active_container_mapping: List = active_term.container if active_term else []
    compacted_item: JsonObject

    # 2)
    if element is None or is_scalar(element):
        return element

    # 3)
    if isinstance(element, List):
        # 3.1)
        result = result_list = []
        # 3.2)
        for item in element:
            # 3.2.1)
            compacted_item = compaction(
                    active_context, active_property, item, compact_arrays, ordered)
            # 3.2.2)
            if compacted_item is not None:
                result.append(compacted_item)
        # 3.3)
        if (len(result_list) == 0 or len(result_list) > 1
                or not compact_arrays or active_property in {GRAPH, SET}
                or LIST in active_container_mapping or SET in active_container_mapping):
            return result_list
        # 3.4)
        else:
            return result_list[0]

    # 4)
    assert isinstance(element, Dict)

    # 5)
    if not active_context._propagate and active_context.previous_context:
        if VALUE not in element:
            if ID not in element or len(element) != 1:
                active_context = active_context.previous_context

    # 6)
    if active_term and active_term.has_local_context:
        active_context = active_term.get_local_context(active_context)

    # 7)
    if VALUE in element or ID in element:
        value_result: JsonObject = value_compaction(active_context, active_term, element)
        if is_scalar(value_result) or active_term and (active_term.type_mapping == JSON):
            return value_result

    # 8)
    if isinstance(element, Dict) and LIST in element and LIST in active_container_mapping:
        return compaction(active_context, active_property, element[LIST], compact_arrays, ordered)

    # 9)
    inside_reverse: bool = active_property == REVERSE

    # 10)
    result = result_map = {}

    # 11)
    # TODO: [5f594ac8] spec improvement; keep this and reuse in step 12.2)
    compacted_types: List[str] = []
    current_active_context: Context = active_context
    if TYPE in element:
        for expanded_type in as_list(element[TYPE]):
            compacted_types.append(
                    iri_compaction(type_scoped_context, cast(str, expanded_type)))
        for ctype in sorted(compacted_types):
            # 11.1)
            typeterm: Optional[Term] = type_scoped_context.terms.get(ctype)
            if typeterm and typeterm.has_local_context:
                active_context = typeterm.get_local_context(current_active_context, False)

    # 12)
    for expanded_property, expanded_value in cast(Dict[str, JsonObject], element).items():
        compacted_value: Optional[str] = None
        # TODO: [5f594ad1] spec improvement; do this once
        alias: str = iri_compaction(active_context, expanded_property, expanded_value)
        alias_term: Optional[Term] = active_context.terms.get(alias)
        alias_container_mapping: List = alias_term.container if alias_term else []
        as_array: bool

        # 12.1)
        if expanded_property == ID:
            # 12.1.1)
            if isinstance(expanded_value, str):
                compacted_value = shorten_iri(active_context, expanded_value)
            # 12.1.2)
            # (see 5f594ad1)
            # 12.1.3)
            result_map[alias] = cast(JsonObject, compacted_value)
            continue

        # 12.2)
        elif expanded_property == TYPE:
            # 12.2.1)
            # TODO: spec improvement (see 5f594ac8)
            # 12.2.2)
                # 12.2.2.1)
                # 12.2.2.2)
                    # 12.2.2.2.1)
                    # 12.2.2.2.2)
            # 12.2.3)
            # (see 5f594ad1)
            # 12.2.4)
            as_array = (active_context._processing_mode == JSONLD11
                        and SET in alias_container_mapping) or not compact_arrays
            # 12.2.5)
            # TODO: spec problem; says to pick first if isinstance(expanded_value, str),
            # but len is 1 is what works...
            add_value(result_map, alias,
                    compacted_types[0] if len(compacted_types) == 1 else compacted_types,
                    as_array)
            # 12.2.6)
            continue

        # 12.3)
        elif expanded_property == REVERSE:
            # 12.3.1)
            compacted_map: JsonMap = cast(JsonMap, compaction(
                    active_context, REVERSE, expanded_value, compact_arrays, ordered))
            # 12.3.2)
            for prop in cast(List[str], list(compacted_map.keys())):
                value: JsonObject = compacted_map[prop]
                prop_term: Optional[Term] = active_context.terms.get(prop)
                # 12.3.2.1)
                if prop_term and prop_term.is_reverse_property:
                    # 12.3.2.1.1)
                    as_array = (active_context._processing_mode == JSONLD11
                                and SET in prop_term.container) or not compact_arrays
                    # 12.3.2.1.2)
                    add_value(result_map, prop, value, as_array)
                    # 12.3.2.1.3)
                    compacted_map.pop(prop)
            # 12.3.3)
            if len(compacted_map) > 0:
                # 12.3.3.1)
                rev_alias: str = iri_compaction(active_context, REVERSE)
                # 12.3.3.2)
                result_map[rev_alias] = compacted_map
            # 12.3.4)
            continue

        # 12.4)
        elif expanded_property == PRESERVE:
            # 12.4.1)
            # 12.4.2)
            if expanded_value != []:
                result_map[PRESERVE] = compaction(active_context, active_property, expanded_value,
                        compact_arrays, ordered)

        # 12.5)
        elif expanded_property == INDEX and INDEX in active_container_mapping:
            continue

        # 12.6)
        elif expanded_property in {DIRECTION, INDEX, LANGUAGE, VALUE}:
            # 12.6.1)
            # 12.6.2)
            result_map[alias] = expanded_value
            continue

        item_active_property: str
        item_active_term: Optional[Term]
        nest_result: JsonMap
        # 12.7)
        if expanded_value == []:
            # 12.7.1)
            item_active_property = iri_compaction(active_context, expanded_property, expanded_value, vocab=True, reverse=inside_reverse)
            item_active_term = active_context.terms.get(item_active_property)
            # 12.7.2 - 12.7.3)
            nest_result = _get_nest_result(active_context, item_active_term, result_map)
            # 12.7.4)
            add_value_as_list(nest_result, item_active_property, [])

        # 12.8)
        expanded_value = as_list(expanded_value)
        assert isinstance(expanded_value, List)

        for expanded_item in expanded_value:
            # 12.8.1)
            item_active_property = iri_compaction(active_context, expanded_property, expanded_item, vocab=True, reverse=inside_reverse)
            item_active_term = active_context.terms.get(item_active_property)
            # 12.8.2 - 12.8.3)
            nest_result = _get_nest_result(active_context, item_active_term, result_map)
            # 12.8.4)
            container: List[str]
            if item_active_term and item_active_term.container is not None:
                container = item_active_term.container
            else:
                container = []
            # 12.8.5)
            as_array = (SET in container
                        or item_active_property in {GRAPH, LIST}
                       ) or not compact_arrays
            # 12.8.6)
            item_to_compact: JsonObject = expanded_item
            item_is_list_object: bool = False
            item_is_graph_object: bool = False
            if isinstance(expanded_item, Dict):
                if LIST in expanded_item:
                    item_to_compact = expanded_item[LIST]
                    item_is_list_object = True
                if is_graph_object(expanded_item):
                    item_to_compact = expanded_item[GRAPH]
                    item_is_graph_object = True

            compacted_item = compaction(
                    active_context, item_active_property, item_to_compact, compact_arrays, ordered)

            # 12.8.7)
            if item_is_list_object:
                assert isinstance(expanded_item, Dict) # TODO: follows indirectly from if test
                #assert isinstance(compacted_item, Dict)
                # 12.8.7.1)
                if not isinstance(compacted_item, List):
                    compacted_item = [compacted_item]
                # 12.8.7.2)
                if LIST not in container:
                    # 12.8.7.2.1)
                    compacted_item = {iri_compaction(active_context, LIST): compacted_item}
                    assert isinstance(compacted_item, Dict) # TODO: redundant, for transpile
                    # 12.8.7.2.2)
                    if INDEX in expanded_item:
                        compacted_item[iri_compaction(active_context, INDEX)] = expanded_item[INDEX]
                    # 12.8.7.2.3)
                    add_value(nest_result, item_active_property, compacted_item, as_array)
                    # TODO: spec problem; does not continue here (thus adding twice!)
                    continue
                # 12.8.7.3)
                elif active_property:
                    nest_result[active_property] = compacted_item

            # 12.8.8)
            map_object: JsonMap
            map_key: Optional[str] = None
            if item_is_graph_object:
                assert isinstance(expanded_item, Dict) # TODO: follows indirectly from if test
                # 12.8.8.1)
                if GRAPH in container and ID in container:
                    # 12.8.8.1.1)
                    map_object = cast(JsonMap, nest_result.setdefault(item_active_property, {}))
                    # 12.8.8.1.2)
                    id_or_none: str = cast(str, expanded_item[ID]) if ID in expanded_item else NONE
                    map_key = iri_compaction(active_context, id_or_none, None, vocab=ID not in expanded_item)
                    # 12.8.8.1.3)
                    add_value(map_object, map_key, compacted_item, as_array)
                # 12.8.8.2)
                elif GRAPH in container and INDEX in container and is_simple_graph_object(expanded_item):
                    # 12.8.8.2.1)
                    map_object = cast(JsonMap, nest_result.setdefault(item_active_property, {}))
                    # 12.8.8.2.2)
                    index_or_none: str = cast(str, expanded_item[INDEX]) if INDEX in expanded_item else NONE
                    map_key = iri_compaction(active_context, index_or_none)
                    # 12.8.8.2.3)
                    add_value(map_object, map_key, compacted_item, as_array)
                # 12.8.8.3)
                elif GRAPH in container and is_simple_graph_object(expanded_item):
                    # 12.8.8.3.1)
                    if isinstance(compacted_item, List) and len(compacted_item) > 1:
                        compacted_item = {iri_compaction(active_context, INCLUDED): compacted_item}
                    # 12.8.8.3.2)
                    add_value(nest_result, item_active_property, compacted_item, as_array)
                # 12.8.8.4)
                else:
                    # 12.8.8.4.1)
                    compacted_item = {iri_compaction(active_context, GRAPH): compacted_item}
                    assert isinstance(compacted_item, Dict)
                    # 12.8.8.4.2)
                    if ID in expanded_item:
                        compacted_item[iri_compaction(active_context, ID)] = shorten_iri(active_context, cast(str, expanded_item[ID]))
                    # 12.8.8.4.3)
                    if INDEX in expanded_item:
                        compacted_item[iri_compaction(active_context, INDEX)] = iri_compaction(active_context, cast(str, expanded_item[INDEX]))
                    # 12.8.8.4.4)
                    add_value(nest_result, item_active_property, compacted_item, as_array)

            # 12.8.9)
            elif (LANGUAGE in container
                  or INDEX in container
                  or ID in container
                  or TYPE in container) and GRAPH not in container:
                assert isinstance(expanded_item, Dict)
                # 12.8.9.1)
                map_object = cast(JsonMap, nest_result.setdefault(item_active_property, {}))
                # 12.8.9.2)
                container_kw: str = cast(str, LANGUAGE if LANGUAGE in container else
                                              INDEX if INDEX in container else
                                              ID if ID in container else
                                              TYPE if TYPE in container else None)
                container_key: str = iri_compaction(active_context, container_kw)
                # 12.8.9.3)
                index_key: str = item_active_term.index if item_active_term and item_active_term.index is not None else INDEX
                # 12.8.9.4)
                if LANGUAGE in container and VALUE in expanded_item:
                    # TODO: spec problem; doesn't mention if compacted_item is scalar
                    if isinstance(compacted_item, Dict):
                        compacted_item = compacted_item[VALUE]
                    map_key = cast(str, expanded_item.get(LANGUAGE))
                # 12.8.9.5)
                elif INDEX in container and index_key == INDEX:
                    # TODO: spec problem? This step is not in the spec!
                    if isinstance(compacted_item, Dict) and VALUE in compacted_item and len(compacted_item) == 1:
                        compacted_item = compacted_item[VALUE]
                    map_key = cast(str, expanded_item.get(INDEX))
                # 12.8.9.6)
                elif INDEX in container and index_key != INDEX:
                    # 12.8.9.6.1)
                    container_key = iri_compaction(active_context, index_key)
                    # 12.8.9.6.2)
                    # 12.8.9.6.3)
                    if isinstance(compacted_item, Dict) and container_key in compacted_item:
                        map_key = _get_map_key_and_deplete_container_key(compacted_item, container_key, as_array)
                # 12.8.9.7)
                elif ID in container:
                    assert isinstance(compacted_item, Dict) # TOOD: check?
                    map_key = cast(Optional[str], compacted_item.pop(container_key, None))
                # 12.8.9.8)
                elif TYPE in container:
                    # 12.8.9.8.1)
                    # 12.8.9.8.2)
                    # 12.8.9.8.3)
                    assert isinstance(compacted_item, Dict) # TOOD: check?
                    if container_key in compacted_item:
                        map_key = _get_map_key_and_deplete_container_key(compacted_item, container_key, as_array)
                    # 12.8.9.8.4)
                    if len(compacted_item) == 1:
                        for key, idval in cast(Dict[str, object], compacted_item).items():
                            if active_context.expand_vocab_iri(key) == ID:
                                compacted_item = compaction(
                                        active_context, item_active_property, {ID: idval})
                # 12.8.9.9)
                if map_key is None:
                    map_key = iri_compaction(active_context, NONE)
                # 12.8.9.10)
                add_value(map_object, map_key, compacted_item, as_array)

            # 12.8.10)
            else:
                add_value(nest_result, item_active_property, compacted_item, as_array)

    # 13)
    return result


def _get_map_key_and_deplete_container_key(
        compacted_item: JsonMap,
        container_key: str,
        as_array: bool) -> Optional[str]:
    container_key_values: List[str] = as_list(compacted_item[container_key])

    key_value: JsonObject = container_key_values.pop(0)
    if not isinstance(key_value, str):
        return None
    map_key: str = key_value

    if compacted_item[container_key] is not container_key_values:
        for remaining in list(container_key_values):
            add_value(compacted_item, container_key, remaining)

    if len(container_key_values) == 1 and not as_array:
        compacted_item[container_key] = container_key_values[0]
    elif len(container_key_values) == 0:
        del compacted_item[container_key]

    return map_key


def _get_nest_result(active_context: Context,
        item_active_term: Optional[Term],
        result_map: JsonMap) -> JsonMap:
    # TODO: spec redundancy, steps are the same...
    # 12.7.2 = 12.8.2)
    nest_result: JsonMap
    if item_active_term and item_active_term.nest_value is not None:
        # 12.7.2.1 = 12.8.2.1)
        if item_active_term.nest_value != NEST \
                and active_context.expand_vocab_iri(item_active_term.nest_value) != NEST:
            raise InvalidNestValueError(item_active_term.nest_value)
        # 12.7.2.2 = 12.8.2.2)
        if item_active_term.nest_value not in result_map:
            result_map[item_active_term.nest_value] = {}
        # 12.7.2.3 = 12.8.2.3)
        return cast(JsonMap, result_map[item_active_term.nest_value])
    # 12.7.3 = 12.8.3)
    else:
        return result_map


def maybe_iri_compaction(active_context: Context,
        iri: Optional[str],
        value: Optional[JsonObject] = None,
        vocab = True,
        reverse = False
        ) -> Optional[str]:
    # 1)
    if iri is None:
        return None
    return iri_compaction(active_context, iri, value, vocab, reverse)


# TODO: spec improvement: allow this to keep IRI:s as is with a flag (also much simpler)!
# Case: To make convenient JSON with preserved input IRIs (or even expanded full IRIs!).
def shorten_iri(active_context: Context, iri: str) -> str:
    return iri_compaction(active_context, iri, None, vocab=False)


def iri_compaction(active_context: Context, iri: str,
        value: Optional[JsonObject] = None,
        vocab = True,
        reverse = False
        ) -> str:
    # 3)
    inverse_context: Dict = get_inverse_context(active_context)

    # 4)
    if vocab and iri in inverse_context:
        # 4.1 + 4.1.2))
        default_language: str = active_context.default_language if active_context.default_language else NONE
        if active_context.default_base_direction is not None:
            # 4.1.1)
            default_language = f'{default_language}_{active_context.default_base_direction}'

        # 4.2)
        if isinstance(value, Dict) and PRESERVE in value:
            values: List = as_list(value[PRESERVE]) # TODO: as_list not needed?
            value = values[0]

        # 4.3)
        containers: List[str] = []

        # 4.4)
        type_or_language: str = LANGUAGE
        type_or_language_value: str = NULL

        # 4.5)
        if isinstance(value, Dict) and INDEX in value and GRAPH not in value:
            containers.append(INDEX)
            containers.append(f'{INDEX}{SET}')

        # 4.6)
        if reverse:
            type_or_language = TYPE
            type_or_language_value = REVERSE
            containers.append(SET)
        # 4.7)
        elif isinstance(value, Dict) and LIST in value:
            # 4.7.1)
            if INDEX not in value:
                containers.append(LIST)
            # 4.7.2)
            valuelist: List[JsonMap] = cast(List[JsonMap], value[LIST])
            # 4.7.3)
            common_type: Optional[str] = None
            common_language: Optional[str] = None
            if len(valuelist) == 0:
                common_language = default_language
            # 4.7.4)
            for item in valuelist:
                # TODO: always expect isinstance(item. Dict)?
                # 4.7.4.1)
                item_language: str = NONE
                item_type: str = NONE
                # 4.7.4.2)
                if isinstance(item, Dict) and VALUE in item:
                    # 4.7.4.2.1)
                    if DIRECTION in item:
                        item_language = f"{cast(str, item.get(LANGUAGE, ''))}_{item[DIRECTION]}"
                    # 4.7.4.2.2)
                    elif LANGUAGE in item:
                        item_language = cast(str, item[LANGUAGE])
                    # 4.7.4.2.3)
                    elif TYPE in item:
                        item_type = cast(str, item[TYPE])
                    # 4.7.4.2.4)
                    else:
                        item_language = NULL
                # 4.7.4.3)
                else:
                    item_type = ID
                # 4.7.4.4)
                if common_language is None:
                    common_language = item_language
                # 4.7.4.5)
                elif item_language != common_language and isinstance(item, Dict) and VALUE in item:
                    common_language = NONE
                # 4.7.4.6)
                if common_type is None:
                    common_type = item_type
                # 4.7.4.7)
                elif item_type != common_type:
                    common_type = NONE
                # 4.7.4.8)
                if common_language == NONE and common_type == NONE:
                    break
            # 4.7.5)
            if common_language is None:
                common_language = NONE
            # 4.7.6)
            if common_type is None:
                common_type = NONE
            # 4.7.7)
            if common_type != NONE:
                type_or_language = TYPE
                type_or_language_value = common_type
            # 4.7.8)
            else:
                type_or_language_value = common_language
        # 4.8)
        elif isinstance(value, Dict) and GRAPH in value:
            # 4.8.1)
            if INDEX in value:
                containers.append(f'{GRAPH}{INDEX}')
                containers.append(f'{GRAPH}{INDEX}{SET}')
            # 4.8.2)
            if ID in value:
                containers.append(f'{GRAPH}{ID}')
                containers.append(f'{GRAPH}{ID}{SET}')
            # 4.8.3)
            containers.append(GRAPH)
            containers.append(f'{GRAPH}{SET}')
            containers.append(SET)
            # 4.8.4)
            if INDEX not in value:
                containers.append(f'{GRAPH}{INDEX}')
                containers.append(f'{GRAPH}{INDEX}{SET}')
            # 4.8.5)
            if ID not in value:
                containers.append(f'{GRAPH}{ID}')
                containers.append(f'{GRAPH}{ID}{SET}')
            # 4.8.6)
            containers.append(INDEX)
            containers.append(f'{INDEX}{SET}')
            # 4.8.7)
            type_or_language = TYPE
            type_or_language_value = ID
        # 4.9)
        else:
            # 4.9.1)
            if isinstance(value, Dict) and VALUE in value:
                # 4.9.1.1)
                if DIRECTION in value and INDEX not in value:
                    type_or_language_value = f"{cast(str, value.get(LANGUAGE, ''))}_{value[DIRECTION]}"
                    containers.append(LANGUAGE)
                    containers.append(f'{LANGUAGE}{SET}')
                # 4.9.1.2)
                elif LANGUAGE in value and INDEX not in value:
                    type_or_language_value = cast(str, value[LANGUAGE])
                    containers.append(LANGUAGE)
                    containers.append(f'{LANGUAGE}{SET}')
                # 4.9.1.3)
                elif TYPE in value:
                    type_or_language_value = cast(str, value[TYPE])
                    type_or_language = TYPE
            # 4.9.2)
            else:
                type_or_language = TYPE
                type_or_language_value = ID
                containers.append(ID)
                containers.append(f'{ID}{SET}')
                containers.append(TYPE)
                containers.append(f'{SET}{TYPE}')
            # 4.9.3)
            containers.append(SET)

        # 4.10)
        containers.append(NONE)
        # 4.11)
        if active_context._processing_mode != JSONLD10:
            if not isinstance(value, Dict) or INDEX not in value:
                containers.append(INDEX)
                containers.append(f'{INDEX}{SET}')
        # 4.12)
        if active_context._processing_mode != JSONLD10:
            if isinstance(value, Dict) and len(value) == 1 and VALUE in value:
                containers.append(LANGUAGE)
                containers.append(f'{LANGUAGE}{SET}')
        # 4.13)
        if type_or_language_value is None:
            type_or_language_value = NULL
        # 4.14)
        preferred_values: List[str] = []
        # 4.15)
        if type_or_language_value == REVERSE:
            preferred_values.append(REVERSE)
        # 4.16)
        if type_or_language_value in {ID, REVERSE} and isinstance(value, Dict) and ID in value:
            # 4.16.1)
            compact_id: str = iri_compaction(active_context, cast(str, value[ID]))
            id_term: Optional[Term] = active_context.terms.get(compact_id)
            if id_term and id_term.iri == value[ID]:
                preferred_values.append(VOCAB)
                preferred_values.append(ID)
                preferred_values.append(NONE)
            # 4.16.2)
            else:
                preferred_values.append(ID)
                preferred_values.append(VOCAB)
                preferred_values.append(NONE)
        # 4.17)
        else:
            preferred_values.append(type_or_language_value)
            preferred_values.append(NONE)
            if isinstance(value, Dict) and LIST in value:
                listvalue: List = cast(List, value[LIST])
                if len(listvalue) == 0:
                    type_or_language = ANY
        # 4.18)
        preferred_values.append(ANY)
        # 4.19)
        for pv in cast(List[str], list(preferred_values)):
            idx: int = pv.find('_')
            if idx > -1:
                preferred_values.append(pv[idx:])
        # 4.20)
        term_key: Optional[str] = term_selection(active_context, iri, containers, type_or_language, preferred_values)
        # 4.21)
        if term_key is not None:
            return term_key

    # 5)
    if vocab and active_context.vocabulary_mapping is not None:
        # 5.1)
        if iri.startswith(active_context.vocabulary_mapping):
            suffix: str = iri[len(active_context.vocabulary_mapping):]
            if len(suffix) > 0 and suffix not in active_context.terms:
                return suffix

    # 6)
    compact_iri: Optional[str] = None

    # 7)
    for key, term_dfn in active_context.terms.items():
        # 7.1)
        if term_dfn.iri is None or term_dfn.iri == iri or not iri.startswith(term_dfn.iri) or not term_dfn.is_prefix:
            continue
        # 7.2)
        candidate: str = f'{key}:{iri[len(term_dfn.iri):]}'
        # 7.3)
        if compact_iri is None or len(candidate) <= len(compact_iri) and candidate < compact_iri:
            if candidate not in active_context.terms or active_context.terms[candidate].iri == iri and value is None:
                compact_iri = candidate

    # 8)
    if compact_iri is not None:
        return compact_iri

    # 9)
    colonx: int = iri.find(':')
    if colonx > -1 and '//' not in iri:
        term: Optional[Term] = active_context.terms.get(iri[0:colonx])
        if term and term.is_prefix:
            raise IRIConfusedWithPrefixError(str(iri))

    # 10)
    if not vocab:
        iri = relativise_iri(active_context.base_iri, iri)

    # 11)
    return iri


def term_selection(active_context: Context,
        keyword_or_iri: str,
        containers: List[str],
        type_or_language: str,
        preferred_values: List[str]) -> Optional[str]:
    # 1)
    # 2)
    inverse_context: Dict = get_inverse_context(active_context)
    # 3)
    container_map: JsonMap = cast(JsonMap, inverse_context[keyword_or_iri])
    # 4)
    for container in containers:
        # 4.1)
        if container not in container_map:
            continue
        # 4.2)
        typelanguage_map: JsonMap = cast(JsonMap, container_map[container])
        # 4.3)
        value_map: Dict[str, str] = cast(Dict[str, str], typelanguage_map[type_or_language])
        # 4.4)
        for item in preferred_values:
            # 4.4.1)
            if item not in value_map:
                continue
            # 4.4.2)
            return value_map[item]
    # 5)
    return None


def value_compaction(
        active_context: Context,
        active_term: Optional[Term],
        value: JsonMap
        ) -> JsonObject:
    # 1)
    result: JsonObject = dict(value)
    # 2)
    # 3)
    inverse_context: Dict = get_inverse_context(active_context)
    # 4)
    language: Optional[str] = active_context.default_language
    if active_term and active_term.language is not None:
        language = active_term.language
    # 5)
    direction: Optional[str] = active_context.default_base_direction
    if active_term and active_term.direction is not None:
        direction = active_term.direction
    # 6)
    if ID in value and len(value) == 1 or len(value) == 2 and INDEX in value:
        # 6.1)
        if active_term and active_term.type_mapping == ID:
            result = shorten_iri(active_context, cast(str, value[ID]))
        # 6.2)
        elif active_term and active_term.type_mapping == VOCAB:
            result = iri_compaction(active_context, cast(str, value[ID]))
    # 7)
    elif active_term and TYPE in value and value[TYPE] == active_term.type_mapping:
        result = value[VALUE]
    # 8)
    elif (active_term and active_term.type_mapping == NONE) or TYPE in value and (active_term is None or value[TYPE] != active_term.type_mapping):
        # 8.1)
        if isinstance(result, Dict) and TYPE in result:
            # TODO: spec problem; if type value is an array...
            rtype: Union[str, List[str]] = result[TYPE]
            if isinstance(rtype, List):
                ctypes: List[str] = []
                for t in rtype:
                    ctypes.append(iri_compaction(active_context, cast(str, t)))
                result[TYPE] = ctypes
            else:
                result[TYPE] = iri_compaction(active_context, cast(str, rtype))
    # 9)
    elif VALUE in value and not isinstance(value[VALUE], str):
        # 9.1)
        if (INDEX in value and active_term and INDEX in active_term.container) or INDEX not in value:
            result = value[VALUE]
    # 10) # TODO: opague spec text.
    elif VALUE in value and (LANGUAGE in value and value[LANGUAGE] == language or language in NULLS and LANGUAGE not in value):
        if DIRECTION not in value or direction is not None and value[DIRECTION] == direction:
            # 10.1)
            if INDEX not in value or (active_term and INDEX in active_term.container):
                result = value[VALUE]
    # 11)
    if isinstance(result, Dict):
        compacted: Dict = {}
        for k, v in cast(JsonMap, result).items():
            compacted[iri_compaction(active_context, k, v)] = v
        result = compacted
    # 12)
    return result



from typing import Any
import json
import sys
import argparse

from ..common import load_json, dump_json
from .expansion import expand
from .compaction import compact
from .flattening import flatten


def err(msg): print(msg, file=sys.stderr)

argparser = argparse.ArgumentParser()
argparser.add_argument('source', nargs='*')
argparser.add_argument('-c', '--context', help='Use to compact expanded JSON-LD')
argparser.add_argument('-e', '--expand-context', help='Use to expand plain JSON to JSON-LD')
argparser.add_argument('-b', '--base', help='Set the base IRI (default is current source)')
argparser.add_argument('-f', '--flatten', action='store_true')

args = argparser.parse_args()

context_data = None
if args.context:
    context_data = load_json(args.context)

doc_paths = args.source or ['-']

for doc_path in doc_paths:
    if len(doc_paths) > 1:
        err(f"Parsing file: '{doc_path}'")

    if doc_path == '-':
        doc_path = '/dev/stdin'
        data = json.load(sys.stdin)
    else:
        data = load_json(doc_path)

    ordered = True
    base_iri = args.base if args.base else f'file://{doc_path}'

    try:
        result: Any = expand(data, base_iri, args.expand_context, ordered=ordered)
        if args.flatten:
            result = flatten(result, ordered=ordered)
        if context_data:
            result = compact(context_data, result, base_iri, ordered=ordered)

        print(dump_json(result, pretty=True))
    except Exception as e:
        err(f"Error in file '{doc_path}'")
        import traceback
        traceback.print_exc()



from typing import List, Dict, Set, Iterable, Iterator, Union, Optional, NamedTuple, cast
from ..common import dump_canonical_json, parse_json
from .base import *
from .context import InvalidBaseDirectionError
from .expansion import InvalidLanguageTaggedStringError
from .flattening import BNodes, NodeMap, make_node_map
from ..rdfterms import (
    RDF_TYPE,
    RDF_VALUE,
    RDF_LIST,
    RDF_FIRST,
    RDF_REST,
    RDF_NIL,
    RDF_DIRECTION,
    RDF_LANGUAGE,
    RDF_JSON,
    RDF_LANGSTRING,
    XSD_BOOLEAN,
    XSD_DOUBLE,
    XSD_INTEGER,
    XSD_STRING,
    I18N,
)


MAX_INT: float = pow(10, 21)

COMPOUND_LITERAL: str = 'compound-literal'
I18N_DATATYPE: str = 'i18n-datatype'

USAGES: str = 'usages'


# TODO: global flag or add to processor
processing_mode: str = JSONLD11


class RdfLiteral(NamedTuple):
    value: str
    datatype: Optional[str] = None
    language: Optional[str] = None


RdfObject = Union[str, RdfLiteral]


class RdfTriple(NamedTuple):
    s: str
    p: str
    o: RdfObject


class RdfGraph:
    name: Optional[str]
    triples: List[RdfTriple]

    def __init__(self, name: Optional[str] = None):
        self.name = name
        self.triples = []

    def add(self, triple: RdfTriple):
        self.triples.append(triple)


class RdfDataset:
    default_graph: RdfGraph
    named_graphs: Dict[str, RdfGraph]

    def __init__(self):
        self.default_graph = RdfGraph()
        self.named_graphs = {}

    def add(self, graph: RdfGraph):
        if graph.name:
            self.named_graphs[graph.name] = graph
        else:
            self.default_graph = graph

    def __iter__(self) -> Iterator[RdfGraph]:
        yield self.default_graph
        yield from self.named_graphs.values()


class _Usage(NamedTuple):
    node: JsonMap
    property: str
    value: JsonMap


def to_rdf_dataset(data: JsonObject,
        rdf_direction: Optional[str] = None) -> RdfDataset:
    bnodes: BNodes = BNodes()
    dataset: RdfDataset = RdfDataset()
    node_map: NodeMap = {DEFAULT: {}}
    make_node_map(bnodes, data, node_map)

    jsonld_to_rdf_dataset(node_map, dataset, bnodes, rdf_direction)

    return dataset


def jsonld_to_rdf_dataset(node_map: NodeMap, dataset: RdfDataset,
                   bnodes: BNodes, rdf_direction: Optional[str] = None):
    # 1)
    for graph_name in cast(List[str], sorted(node_map.keys())):
        graph: Dict = node_map[graph_name]
        # 1.1)
        if not (is_iri(graph_name) or is_blank(graph_name) or graph_name == DEFAULT):
            continue
        # 1.2)
        triples: RdfGraph
        if graph_name == DEFAULT:
            triples = dataset.default_graph
        else:
            triples = RdfGraph(graph_name)
            dataset.add(triples)
        # 1.3)
        for subject in cast(Iterable[str], sorted(graph.keys())):
            node: JsonMap = graph[subject]
            # 1.3.1)
            if not is_iri_or_blank(subject):
                continue
            # 1.3.2)
            for property in cast(Iterable[str], sorted(node.keys())):
                # TODO: isn't this always a list if this is properly expanded
                # and flattened (as this algoritm implicitly expects..)?
                values: List[object] = as_list(node[property])
                # 1.3.2.1)
                if property == TYPE:
                    for type in values:
                        triples.add(RdfTriple(subject, RDF_TYPE, cast(str, type)))
                # 1.3.2.2)
                elif property in KEYWORDS:
                    continue
                # 1.3.2.3)
                elif is_blank(property): # TODO: and not options.produce_generalized_rdf
                    continue
                # 1.3.2.4)
                elif not is_iri_or_blank(property):
                    continue
                # 1.3.2.5)
                else:
                    for item in values:
                        assert isinstance(item, Dict)
                        # 1.3.2.5.1)
                        list_triples: List[RdfTriple] = []
                        # 1.3.2.5.2)
                        rdf_object: RdfObject = cast(RdfObject,
                                object_to_rdf_data(item, list_triples, bnodes, rdf_direction))
                        list_triples.append(RdfTriple(subject, property, rdf_object))
                        # 1.3.2.5.3)
                        for triple in list_triples:
                            triples.add(triple)


def object_to_rdf_data(item: JsonMap, list_triples: List, bnodes: BNodes,
        rdf_direction: Optional[str] = None) -> Optional[RdfObject]:
    # 1)
    if ID in item and not is_iri_or_blank(cast(str, item[ID])):
        return None
    # 2)
    if ID in item:
        return cast(str, item[ID])
    # 3)
    if LIST in item:
        return list_to_rdf_list(cast(List, item[LIST]), list_triples, bnodes, rdf_direction)
    # 4)
    assert VALUE in item
    value: JsonObject = item[VALUE]
    # 5)
    datatype: Optional[str] = cast(Optional[str], item.get(TYPE))
    # 6)
    if datatype is not None and not is_iri(datatype) and not datatype == JSON:
        return None
    # 7)
    if LANGUAGE in item and not is_lang_tag(cast(str, item[LANGUAGE])):
        return None
    # 8)
    if datatype == JSON:
        value = dump_canonical_json(value)
        datatype = RDF_JSON
    # 9)
    if isinstance(value, bool):
        value = 'true' if value else 'false'
        if datatype is None:
            datatype = XSD_BOOLEAN
    # 10)
    elif isinstance(value, (int, float)) and ((cast(float, value) % 1 > 0 or cast(float, value) >= MAX_INT) or datatype == XSD_DOUBLE):
        value = str(value) # TODO: to_canonical_double(value)
        if datatype is None:
            datatype = XSD_DOUBLE
    # 11)
    elif isinstance(value, (int, float)):
        value = str(value) # TODO: to_canonical_integer(value)
        if datatype is None:
            datatype = XSD_INTEGER
    # 12)
    elif datatype is None:
        datatype = RDF_LANGSTRING if LANGUAGE in item else XSD_STRING
    # 13)
    assert isinstance(value, str)
    literal: Optional[RdfObject] = None
    if DIRECTION in item and rdf_direction is not None:
        # 13.1)
        language: str = cast(str, item.get(LANGUAGE, ''))
        language = language.lower()
        # 13.2)
        if rdf_direction == I18N_DATATYPE:
            datatype = f'{I18N}{language}_{item[DIRECTION]}'
            literal = RdfLiteral(value, datatype)
        # 13.3)
        elif rdf_direction == COMPOUND_LITERAL:
            # 13.3.1)
            nodeid: str = bnodes.make_bnode_id()
            literal = nodeid
            # 13.3.2)
            # TODO: spec errata? says to use item[VALUE] (not to use the stringified value)
            list_triples.append(RdfTriple(nodeid, RDF_VALUE, RdfLiteral(value)))
            # 13.3.3)
            if LANGUAGE in item:
                list_triples.append(RdfTriple(nodeid, RDF_LANGUAGE, RdfLiteral(cast(str, item[LANGUAGE]))))
            # 13.3.4)
            list_triples.append(RdfTriple(nodeid, RDF_DIRECTION, RdfLiteral(cast(str, item[DIRECTION]))))
    # 14)
    else:
        literal = RdfLiteral(value, datatype, cast(Optional[str], item.get(LANGUAGE)))
    # 15)
    return literal


def list_to_rdf_list(l: List[JsonMap], list_triples: List, bnodes: BNodes,
        rdf_direction: Optional[str] = None) -> str:
    # 1)
    if len(l) == 0:
        return RDF_NIL
    # 2)
    # NOTE: made this a little different from spec...
    # 3)
    subject: str = bnodes.make_bnode_id()
    first: str = subject
    i: int = 0
    for item in l:
        # 3.1)
        embedded_triples: List = []
        # 3.2)
        obj: Optional[RdfObject] = object_to_rdf_data(item, embedded_triples, bnodes, rdf_direction)
        # 3.3)
        if obj is not None:
            list_triples.append(RdfTriple(subject, RDF_FIRST, obj))
        # 3.4)
        i += 1
        next_subject: str = bnodes.make_bnode_id() if i < len(l) else RDF_NIL
        list_triples.append(RdfTriple(subject, RDF_REST, next_subject))
        subject = next_subject
        # 3.5)
        list_triples += embedded_triples
    # 4)
    return first if i > 0 else RDF_NIL


def to_jsonld(dataset: RdfDataset,
        ordered=False,
        rdf_direction: Optional[str] = None,
        use_native_types=False,
        use_rdf_type=False) -> List[JsonMap]:
    # 1)
    default_graph: Dict[str, JsonMap] = {}
    # 2)
    graph_map: Dict[str, Dict[str, JsonMap]] = {DEFAULT: default_graph}
    # 3)
    referenced_once: Dict[str, Union[_Usage, bool]] = {}
    # 4)
    compound_literal_subjects: Dict[str, Set[str]] = {}

    # 5)
    for graph in dataset:
        # 5.1)
        name: str = DEFAULT if graph.name is None else graph.name
        # 5.2)
        graph_map.setdefault(name, {})
        # 5.3)
        compound_literal_subjects.setdefault(name, set())
        # 5.4)
        if name != DEFAULT:
            default_graph.setdefault(name, {ID: name})
        # 5.5)
        node_map: Dict[str, JsonMap] = cast(Dict[str, JsonMap], graph_map[name])
        # 5.6)
        compounds: Set[str] = compound_literal_subjects[name]

        # 5.7)
        for triple in graph.triples:
            # 5.7.1)
            if triple.s not in node_map:
                node_map[triple.s] = {ID: triple.s}
            # 5.7.2)
            node: JsonMap = node_map[triple.s]
            # 5.7.3)
            if rdf_direction == COMPOUND_LITERAL and triple.p == RDF_DIRECTION:
                compounds.add(triple.s)
            # 5.7.4)
            if isinstance(triple.o, str) and triple.o not in node_map:
                node_map[triple.o] = {ID: triple.o}
            # 5.7.5)
            if triple.p == RDF_TYPE and not use_rdf_type and isinstance(triple.o, str):
                types: List = cast(List, node.setdefault(TYPE, []))
                if not any(t == triple.o for t in types):
                    types.append(triple.o)
                continue
            # 5.7.6)
            value: JsonMap = to_jsonld_object(triple.o, rdf_direction, use_native_types)
            # 5.7.7)
            values: List[JsonObject] = cast(List, node.setdefault(triple.p, []))
            # 5.7.8)
            # TODO: spec errata; continue instead, to avoid duplicates in 5.7.9
            # (see fromRdf/0022)
            if any(node_equals(v, value) for v in values):
                continue
            values.append(value)
            # 5.7.9)
            if triple.o == RDF_NIL:
                # 5.7.9.1)
                obj: JsonMap = node_map[cast(str, triple.o)]
                obj_usages: List = cast(List, obj.setdefault(USAGES, []))
                # 5.7.9.2)
                obj_usages.append(_Usage(node, triple.p, value))
            # 5.7.10)
            elif triple.o in referenced_once:
                referenced_once[cast(str, triple.o)] = False
            # 5.7.11)
            elif isinstance(triple.o, str) and is_blank(triple.o):
                # 5.7.11.1)
                usage: _Usage = _Usage(node, triple.p, value)
                referenced_once[triple.o] = usage
                # TODO: spec seems to have an overeager copy/paste error
                # ("and values to the usages array") from just above.

    # 6)
    for name, graph_object in graph_map.items():
        # 6.1)
        if name in compound_literal_subjects:
            # TODO: handle getitem in transpile
            graph_compounds: Set[str] = compound_literal_subjects[name]
            for cl in graph_compounds:
                # 6.1.1)
                cl_entry: object = referenced_once.get(cl)
                if not isinstance(cl_entry, _Usage):
                    continue
                # 6.1.2)
                c_node: JsonMap = cl_entry.node
                # 6.1.3)
                c_property: str = cl_entry.property
                # 6.1.4)
                # TODO: spec errata? not used...
                #c_value: JsonObject = cl_entry.value
                # 6.1.5)
                cl_node: JsonMap = cast(JsonMap, graph_object.pop(cl))
                # 6.1.6)
                for cl_ref in cast(List, c_node[c_property]):
                    assert isinstance(cl_ref, Dict)
                    if not cl_ref[ID] == cl:
                        continue
                    # 6.1.6.1)
                    del cl_ref[ID]
                    # 6.1.6.2)
                    cl_ref[VALUE] = cast(Dict, cast(List, cl_node[RDF_VALUE])[0])[VALUE]
                    # 6.1.6.3)
                    if RDF_LANGUAGE in cl_node:
                        cl_ref[LANGUAGE] = cast(Dict, cast(List, cl_node[RDF_LANGUAGE])[0])[VALUE]
                        if not is_lang_tag(cast(str, cl_ref[LANGUAGE])):
                            raise InvalidLanguageTaggedStringError(str(cl_ref[LANGUAGE]))
                    # 6.1.6.4)
                    if RDF_DIRECTION in cl_node:
                        cl_ref[DIRECTION] = cast(Dict, cast(List, cl_node[RDF_DIRECTION])[0])[VALUE]
                        if not cl_ref[DIRECTION] in DIRECTIONS:
                            raise InvalidBaseDirectionError(str(cl_ref[DIRECTION]))
        # 6.2)
        if RDF_NIL not in graph_object:
            continue
        # 6.3)
        nil: JsonMap = cast(JsonMap, graph_object[RDF_NIL])
        # 6.4)
        if USAGES not in nil:
            continue
        for usage in cast(List[_Usage], nil[USAGES]):
            # 6.4.1)
            unode: JsonMap = usage.node
            uproperty: str = usage.property
            head: JsonMap = usage.value
            # 6.4.2)
            list_values: List = []
            list_nodes: List = []
            # 6.4.3)
            while uproperty == RDF_REST and is_well_formed_list(unode):
                # 6.4.3.3)
                node_usage: Union[_Usage, bool] = referenced_once[cast(str, unode[ID])]
                if not isinstance(node_usage, _Usage):
                    break
                # 6.4.3.1)
                list_values.append(cast(List, unode[RDF_FIRST])[0])
                # 6.4.3.2)
                list_nodes.append(unode[ID])
                # 6.4.3.4)
                unode = node_usage.node
                uproperty = node_usage.property
                head = node_usage.value
                # 6.4.3.5)
                if is_iri(cast(str, unode[ID])):
                    break
            # 6.4.4)
            del head[ID]
            # 6.4.5)
            list_values.reverse()
            # 6.4.6)
            head[LIST] = list_values
            # 6.4.7)
            for node_id in list_nodes:
                del graph_object[node_id]

    # 7)
    result: List[JsonMap] = []
    # 8)
    subjects: List[str] = list(default_graph.keys())
    if ordered:
        subjects.sort()
    for subject in subjects:
        s_node: JsonMap = cast(JsonMap, default_graph[subject])
        # 8.1)
        if subject in graph_map:
            subject_graph: Dict[str, JsonMap] = graph_map[subject]
            # 8.1.1)
            named_graphs: List = []
            s_node[GRAPH] = named_graphs
            # 8.1.2)
            graph_names: List = list(subject_graph.keys())
            if ordered:
                graph_names.sort()
            for graph_name in graph_names:
                named: JsonMap = cast(JsonMap, subject_graph[graph_name])
                if USAGES in named:
                    del named[USAGES]
                if not len(named) == 1 and ID in named:
                    named_graphs.append(named)
        # 8.2)
        if USAGES in s_node:
            del s_node[USAGES]
        if not len(s_node) == 1 and ID in s_node:
            result.append(s_node)
    # 9)
    return result


def to_jsonld_object(value: RdfObject,
        rdf_direction: Optional[str],
        use_native_types: bool) -> JsonMap:
    # 1)
    if isinstance(value, str):
        return {ID: value}

    # 2)
    literal: RdfLiteral = value
    # 2.1)
    result: JsonMap = {}
    # 2.2)
    # TODO: spec errata? says: value
    converted_value: JsonObject = literal.value
    # 2.3)
    rtype: Optional[str] = None

    # 2.4)
    if use_native_types:
        if literal.datatype == XSD_STRING:
            converted_value = literal.value
        elif literal.datatype == XSD_BOOLEAN:
            if literal.value == 'true':
                converted_value = True
            elif literal.value == 'false':
                converted_value = False
            else:
                rtype = XSD_BOOLEAN
        elif literal.datatype == XSD_INTEGER and literal.value.isdecimal():
            converted_value = int(literal.value)
        elif literal.datatype == XSD_DOUBLE:# and all(c == '.' or c.isdecimal() for c in literal.value):
            try:
                converted_value = float(literal.value)
            except ValueError:
                pass
        # TODO: spec errata; only done in an otherwise below (thus fromRdf/0018 fails)
        elif literal.datatype != XSD_STRING:
            rtype = literal.datatype
    # 2.5)
    elif literal.datatype == RDF_JSON and processing_mode != JSONLD10:
        try:
            converted_value = cast(JsonObject, parse_json(literal.value))
        except:# json.decoder.JSONDecodeError
            pass
        rtype = JSON
    # 2.6)
    elif rdf_direction == I18N_DATATYPE and literal.datatype is not None and literal.datatype.startswith(I18N):
        # 2.6.1)
        converted_value = literal.value
        # 2.6.2)
        frag_id: str = literal.datatype[len(I18N):]
        i: int = frag_id.find('_')
        lang: str = frag_id
        direction: str = ''
        if i > -1:
            lang = frag_id[:i]
            direction = frag_id[i + 1:]
        if len(lang) > 0:
            result[LANGUAGE] = lang
        # 2.6.3)
        if len(direction) > 0:
            result[DIRECTION] = direction
    # 2.7)
    elif literal.language is not None:
        result[LANGUAGE] = literal.language
    # 2.8)
    elif literal.datatype and literal.datatype != XSD_STRING:
        rtype = literal.datatype

    # 2.9)
    result[VALUE] = converted_value
    # 2.10)
    if rtype:
        result[TYPE] = rtype

    # 2.11)
    return result


def is_iri_or_blank(iri: str) -> bool:
    return is_iri(iri) or is_blank(iri)


def is_well_formed_list(node: JsonMap) -> bool:
    return is_blank(cast(str, node[ID])) and \
            _has_list_with_one_item(node, RDF_FIRST) and \
            _has_list_with_one_item(node, RDF_REST) and \
            (len(node) == 3 or len(node) == 4 and TYPE in node and
                    as_list(node[TYPE])[0] == RDF_LIST)


def _has_list_with_one_item(node: JsonMap, p) -> bool:
    return p in node and isinstance(node[p], List) and len(cast(List, node[p])) == 1



from typing import Optional, Dict, List, Set, Union, cast
from ..common import warning
from .base import *


NodeMap = Dict[str, JsonMap]


class ConflictingIndexesError(JsonLdError): pass


class BNodes:
    i: int
    id_map: Dict[str, str]

    def __init__(self):
        self.i = 0
        self.id_map = {}

    def make_bnode_id(self, identifier: Optional[str] = None) -> str:
        if identifier in self.id_map:
            return self.id_map[identifier]
        bnode_id: str = f'_:b{str(self.i)}'
        self.i += 1
        if identifier is not None:
            self.id_map[identifier] = bnode_id
        return bnode_id


def flatten(element: JsonObject, ordered=False, bnodes: Optional[BNodes] = None) -> JsonObject:
    if bnodes is None:
        bnodes = BNodes()

    # 1)
    node_map: NodeMap = {DEFAULT: {}}
    # 2)
    make_node_map(bnodes, element, node_map)
    # 3)
    default_graph: Dict = node_map[DEFAULT]
    # 4)
    graph_names: List[str] = list(node_map.keys())
    if ordered:
        graph_names.sort()
    for graph_name in graph_names:
        if graph_name == DEFAULT:
            continue
        graph: JsonMap = cast(JsonMap, node_map[graph_name])
        # 4.1)
        if graph_name not in default_graph:
            default_graph[graph_name] = {ID: graph_name}
        # 4.2)
        entry: JsonMap = cast(JsonMap, default_graph[graph_name])
        # 4.3)
        entry_graph: List[JsonMap] = []
        entry[GRAPH] = entry_graph 
        # 4.4)
        node_ids: List[str] = list(graph.keys())
        if ordered:
            node_ids.sort()
        for node_id in node_ids:
            node: JsonMap = cast(JsonMap, graph[node_id])
            if len(node) > 1 or ID not in node:
                entry_graph.append(node)
    # 5)
    flattened: List[JsonMap] = []
    # 6)
    top_node_ids: List[str] = list(default_graph.keys())
    if ordered:
        top_node_ids.sort()
    for node_id in top_node_ids:
        top_node: JsonMap = cast(JsonMap, default_graph[node_id])
        if len(top_node) > 1 or ID not in top_node:
            flattened.append(top_node)
    # 7)
    return flattened


def make_node_map(bnodes: BNodes,
        element,
        node_map: NodeMap,
        active_graph: str = DEFAULT,
        active_subject: Optional[Union[str, JsonMap]] = None,
        active_property: Optional[str] = None,
        list_map: Optional[Dict[str, JsonObject]] = None):
    # 1)
    if isinstance(element, List):
        for item in element:
            # 1.1)
            make_node_map(bnodes, item, node_map, active_graph, active_subject,
                          active_property, list_map)
        return

    # 2)
    assert isinstance(element, Dict)
    # TODO: spec problem; seems to expect node_map[active_graph] to already be
    # set?
    graph: JsonMap = node_map.setdefault(active_graph, {})
    subject_node: Optional[JsonMap] = None
    if isinstance(active_subject, str):
        subject_node = cast(JsonMap, graph.get(active_subject))

    # 3)
    if TYPE in element:
        etype: object = element[TYPE]
        mapped_types: List = []
        for item in cast(List[str], as_list(etype)):
            # 3.1)
            if is_blank(item):
                item = bnodes.make_bnode_id(item)
            mapped_types.append(item)
        element[TYPE] = mapped_types if isinstance(etype, List) else mapped_types[0]

    # 4)
    if VALUE in element:
        assert subject_node is not None
        assert isinstance(active_property, str)
        # 4.1)
        if list_map is None:
            # 4.1.1)
            if active_property not in subject_node:
                subject_node[active_property] = [element]
            # 4.1.2)
            else:
                elements: List[JsonObject] = cast(List[JsonObject], subject_node[active_property])
                if not any(node_equals(element, el) for el in elements):
                    elements.append(element)
        # 4.2)
        else:
            cast(JsonList, list_map[LIST]).append(element)

    # 5)
    elif LIST in element:
        assert subject_node is not None
        assert isinstance(active_property, str)
        # 5.1)
        result: JsonMap = {LIST: []}
        # 5.2)
        make_node_map(bnodes, element[LIST], node_map, active_graph,
                      active_subject, active_property, result)
        # 5.3)
        if list_map is None:
            cast(JsonList, subject_node[active_property]).append(result)
        # 5.4)
        else:
            cast(JsonList, list_map[LIST]).append(result)

    # 6)
    else:
        # 6.1)
        eid: str
        if ID in element:
            eid = cast(str, element.pop(ID))
            if eid is None or is_blank(eid):
                eid = bnodes.make_bnode_id(eid)
        # 6.2)
        else:
            eid = bnodes.make_bnode_id(None)
        # 6.3)
        if eid not in graph:
            graph[eid] = {ID: eid}
        # 6.4)
        node: JsonMap = cast(JsonMap, graph[eid])
        # 6.5)
        if isinstance(active_subject, Dict): # reverse property relationship
            assert isinstance(active_property, str)
            # 6.5.1)
            if active_property not in node:
                node[active_property] = [active_subject]
            # 6.5.2)
            else:
                subjects: List[JsonMap] = cast(List[JsonMap] , node[active_property])
                if not any(node_equals(active_subject, subj) for subj in subjects):
                    subjects.append(active_subject)

        # 6.6)
        elif active_property is not None:
            assert subject_node is not None
            # 6.6.1)
            reference: JsonMap = {ID: eid}
            # 6.6.2)
            if list_map is None:
                # 6.6.2.1)
                if active_property not in subject_node:
                    subject_node[active_property] = [reference]
                # 6.6.2.2)
                objects: List[JsonMap] = cast(List[JsonMap], subject_node[active_property])
                if reference not in objects:
                    objects.append(reference)
            # 6.6.3)
            else:
                cast(JsonList, list_map[LIST]).append(reference)

        # 6.7)
        if TYPE in element:
            # TODO: spec problem; seems to expect node[TYPE] to already be set?
            ntypes: List = cast(List, node.setdefault(TYPE, []))
            for ntype in as_list(element[TYPE]):
                if ntype not in ntypes:
                    ntypes.append(ntype)
            del element[TYPE]

        # 6.8)
        if INDEX in element:
            if INDEX in node and node[INDEX] != element[INDEX]:
                raise ConflictingIndexesError(str(node[INDEX]))
            node[INDEX] = element.pop(INDEX)

        # 6.9)
        if REVERSE in element:
            # 6.9.1)
            referenced_node: JsonMap = {ID: eid}
            # 6.9.2)
            reverse_map: JsonMap = cast(JsonMap, element[REVERSE])
            # 6.9.3)
            for property, values in reverse_map.items():
                # 6.9.3.1)
                for value in cast(List, values):
                    # 6.9.3.1.1)
                    make_node_map(bnodes, value, node_map, active_graph,
                                  referenced_node, property)
            # 6.9.4)
            del element[REVERSE]

        # 6.10)
        if GRAPH in element:
            make_node_map(bnodes, element[GRAPH], node_map, eid)
            del element[GRAPH]

        # 6.11)
        if INCLUDED in element:
            make_node_map(bnodes, element[INCLUDED], node_map, active_graph)
            del element[INCLUDED]

        # 6.12)
        properties: List[str] = list(element.keys())
        properties.sort()
        for property in properties:
            evalue: JsonObject = element[property]
            # 6.12.1)
            if is_blank(property):
                property = bnodes.make_bnode_id(property)
            # 6.12.2)
            if property not in node:
                node[property] = []
            # 6.12.3)
            make_node_map(bnodes, evalue, node_map, active_graph, eid, property)


def merge_node_maps(node_maps: Dict[str, NodeMap]):
    # 1)
    result: JsonMap = {}
    # 2)
    for graph_name, node_map in node_maps.items():
        for node_id, node in node_map.items():
            # 2.1)
            merged_node: JsonMap = cast(JsonMap,
                    result.setdefault(node_id, {ID: node_id}))
            # 2.2)
            for property, values in node.items():
                # 2.2.1)
                if property != TYPE and property in KEYWORDS:
                    merged_node[property] = values
                # 2.2.2)
                else:
                    # TODO: recursive merge_node_maps...?
                    existing: List = cast(List,
                            merged_node.setdefault(property, []))
                    existing += cast(List, values)
    # 3)
    return result



from typing import Optional, Dict, List, Set, Union, cast
from ..jsonld.base import CONTEXT, GRAPH, ID, TYPE, VOCAB, as_list


ListOrJsonMap = Union[List, Dict[str, object]]


def map_to(target_map: Dict, indata, drop_unmapped=False) -> ListOrJsonMap:
    result: ListOrJsonMap = {} if isinstance(indata, Dict) else []
    _modify(target_map, indata, result, drop_unmapped)
    return result


def _modify(target_map: Dict, ino: ListOrJsonMap, outo: Union[Dict, List], drop_unmapped: bool):
    if isinstance(ino, Dict):
        for k, v in cast(Dict[str, object], ino).items(): # TODO: cast just for transpile
            _modify_pair(target_map, k, v, outo, drop_unmapped)
    elif isinstance(ino, List):
        i: int = 0
        for v in ino:
            _modify_pair(target_map, i, v, outo, drop_unmapped)
            i += 1


def _modify_pair(target_map: Dict, k: Union[str, int], v: object, outo: Union[Dict, List], drop_unmapped: bool):
    mapo: Dict[Union[str, int], Union[List, Dict, str]] = _map(target_map, k, v, drop_unmapped)

    for mapk, mapv in mapo.items():
        outv: Union[List, Dict]
        if isinstance(mapv, List):
            outv = []
            _modify(target_map, mapv, outv, drop_unmapped)
            mapv = outv
        elif isinstance(mapv, Dict):
            outv = {}
            _modify(target_map, mapv, outv, drop_unmapped)
            mapv = outv

        if isinstance(outo, Dict):
            if mapk in outo:
                values: List = as_list(outo[mapk])
                values += as_list(mapv)
                mapv = values

            outo[mapk] = mapv
        else:
            outo.append(mapv)


def _map(target_map: Dict, key: Union[str, int], value, drop_unmapped=False) -> Dict:
    somerule: object = target_map.get(key)

    if drop_unmapped and isinstance(key, str) and key[0] != '@' and somerule is None:
        return {}

    if isinstance(value, List):
        remapped: List[object] = []
        for v in value:
            item: object = target_map[v] if isinstance(v, str) and  v in target_map else v
            if isinstance(item, List):
                remapped += item
            else:
                remapped.append(item)
        value = remapped

    if somerule is None:
        return {key: value}

    out: Dict = {}

    mapped_keypaths: Set[str] = set()

    for rule in as_list(somerule):
        if isinstance(rule, str):
            out[rule] = value
            break

        if isinstance(rule, Dict):
            objectvalues: List[Dict] = value

            property: Optional[str] = rule.get('property')
            property_from: Optional[str] = rule.get('propertyFrom')

            # TODO: use both property and property_from if present
            if property_from is not None:
                first: Dict = objectvalues[0]
                property_from_object: List[Dict] = first[property_from]
                property = property_from_object[0][ID]

            if property in target_map:
                property = as_list(target_map[property])[0]

            outvalue: List[object] = []
            # TODO: if match + use base_map

            value_from: Optional[str] = rule.get('valueFrom')
            if value_from is not None:
                for v in objectvalues:
                    assert isinstance(v, Dict)
                    match: Optional[Dict[str, str]] = rule.get('match')
                    if match is None or TYPE in match and any(
                            t == match[TYPE] for t in cast(List, v[TYPE])):
                        vv: object = v.get(value_from)
                        if isinstance(vv, List):
                            for m in vv:
                                outvalue.append(m)
                        else:
                            outvalue.append(vv)
            else:
                outvalue = value

            #outvalue = [target_map.get(v, v) for v in outvalue]
            mappedvalue: List[object] = []
            for v in outvalue:
                mappedvalue.append(target_map.get(v, v) if isinstance(v, str) else v)
            outvalue = mappedvalue

            if property is not None and outvalue:
                if value_from is not None:
                    mapped_key = f'{key} {value_from}'

                    # Skipping rule if earlier rule used the same source parts
                    if mapped_key in mapped_keypaths:
                        continue

                    mapped_keypaths.add(mapped_key)

                out[property] = outvalue
                # TODO: Using all rules since they may be complementary.
                # Might need to combine those into one rule instead.
                #break

    return out



from typing import NamedTuple, Optional, Tuple, Dict, List, Set, Union, cast

from collections import OrderedDict

from ..jsonld.base import CONTEXT, GRAPH, ID, LIST, REVERSE, TYPE, VOCAB, as_list, JsonMap
from ..jsonld.extras.index import make_index


RDF: str = 'http://www.w3.org/1999/02/22-rdf-syntax-ns#'
RDF_Property: str = f'{RDF}Property'
RDF_Statement: str = f'{RDF}Statement'
RDF_subject: str = f'{RDF}subject'
RDF_predicate: str = f'{RDF}predicate'
RDF_object: str = f'{RDF}object'

RDFS: str = 'http://www.w3.org/2000/01/rdf-schema#'
RDFS_Class: str = f'{RDFS}Class'
RDFS_subClassOf: str = f'{RDFS}subClassOf'
RDFS_subPropertyOf: str = f'{RDFS}subPropertyOf'
RDFS_domain: str = f'{RDFS}domain'
RDFS_range: str = f'{RDFS}range'

OWL: str = 'http://www.w3.org/2002/07/owl#'
OWL_Class: str = f'{OWL}Class'
OWL_Datatype: str = f'{OWL}Datatype'
OWL_ObjectProperty: str = f'{OWL}ObjectProperty'
OWL_DatatypeProperty: str = f'{OWL}DatatypeProperty'
OWL_Restriction: str = f'{OWL}Restriction'
OWL_equivalentClass: str = f'{OWL}equivalentClass'
OWL_equivalentProperty: str = f'{OWL}equivalentProperty'
OWL_inverseOf: str = f'{OWL}inverseOf'
OWL_propertyChainAxiom: str = f'{OWL}propertyChainAxiom'
OWL_onProperty: str = f'{OWL}onProperty'
OWL_hasValue: str = f'{OWL}hasValue'
OWL_allValuesFrom: str = f'{OWL}allValuesFrom'

SKOS: str = 'http://www.w3.org/2004/02/skos/core#'
SKOS_broadMatch: str = f'{SKOS}broadMatch'
SKOS_closeMatch: str = f'{SKOS}closeMatch'
SKOS_exactMatch: str = f'{SKOS}exactMatch'
SKOS_narrowMatch: str = f'{SKOS}narrowMatch'
SKOS_mappingRelation: str = f'{SKOS}mappingRelation'

SYMMETRIC: Set[str] = {
  OWL_equivalentClass,
  OWL_equivalentProperty,
  SKOS_closeMatch,
  SKOS_exactMatch,
}


Candidates = List[Tuple[Optional[str], Dict]]


class BaseRelation(NamedTuple):
    rel: Optional[str]
    base: str
    priority: int = -1


def make_target_map(vocab: object, target: object) -> Dict:
    target_dfn: Dict[str, object] = OrderedDict()
    if isinstance(target, str):
        target_dfn[VOCAB] = target
    else:
        if isinstance(target, Dict):
            target = target[CONTEXT]
        for dfn in as_list(target):
            target_dfn.update(cast(Dict, dfn))

    graph: List[JsonMap] = vocab if isinstance(vocab, List) else cast(List, cast(Dict, vocab)[GRAPH])

    vocab_index: Dict[str, JsonMap] = make_index(graph)

    target_map: Dict[str, object] = {}

    for obj in graph:
        id: Optional[str] = cast(Optional[str], obj[ID])

        _process_class_relations(obj, vocab_index, target_dfn, target_map)

        _process_property_relations(obj, vocab_index, target_dfn, target_map)

        _process_reified_forms(obj, vocab_index, target_map)

    identity_map: Dict[str, str] = {}

    for key, rule in target_map.items():
        rules: List[Tuple[int, Union[Dict, str]]] = sorted(as_list(rule),
                key=lambda it: (
                        cast(int, cast(Tuple, it)[0]),
                        cast(Tuple, it)[1].get('match') is not None
                        if isinstance(cast(Tuple, it)[1], Dict) else False
                    ),
                reverse=True)
        target_map[key] = [it for priority, it in rules] # keep all

        # Add identity mappings to keep target terms in source:
        for priority, it in rules:
            if isinstance(it, str):
                identity_map[it] = it
            break

    target_map.update(identity_map)

    return target_map


def _process_class_relations(obj: Dict, vocab_index: Dict, target: Dict[str, object], target_map: Dict):
    rels: List[str] = [OWL_equivalentClass, RDFS_subClassOf]

    # TODO: rework this even more like process_property_relations
    base_rels: List[BaseRelation] = []

    id: Optional[str] = cast(Optional[str], obj.get(ID))

    id_target_prio: int = 0
    if id is not None:
        id_target_prio = _get_target_priority(target, id)
        if id_target_prio:
            base_rels.append(BaseRelation(None, id, id_target_prio))

    candidates: Candidates = _collect_candidates(obj, rels)

    seen_candidates: Set[str] = set()

    while candidates:
        crel, candidate = candidates.pop(0)
        if ID not in candidate:
            continue
        candidate_id: str = candidate[ID]
        candidate = cast(Dict, vocab_index.get(candidate_id, candidate))

        target_prio: int = _get_target_priority(target, candidate_id)
        if target_prio:
            base_rels.append(BaseRelation(crel, candidate_id, target_prio))
        elif crel in SYMMETRIC and id_target_prio:
            assert id is not None
            _add_rule(target_map, candidate_id, id, id_target_prio)
        elif candidate_id not in seen_candidates:
            _extend_candidates(candidates, candidate, rels)

        seen_candidates.add(candidate_id)

    if id is not None and not id_target_prio:
        base_rels = [it for it in base_rels if it.base != id]

        if len(base_rels) > 0:
            base_classes: List[str] = []
            for baserel in base_rels:
                class_dfn: Optional[Dict] = cast(Optional[Dict], vocab_index.get(baserel.base))
                if class_dfn is None:
                    class_dfn = {ID: baserel.base}
                base_classes.append(cast(str, class_dfn[ID]))
                break
                #if baserel.rel == OWL_equivalentClass:
                #    _add_rule(target_map, baserel.base, id)
            _add_rule(target_map, id, base_classes)


def _process_property_relations(obj: Dict, vocab_index: Dict, target: Dict[str, object], target_map: Dict):
    rels: List[str] = [OWL_equivalentProperty, RDFS_subPropertyOf]

    if ID not in obj: # TODO: OK?
        return

    id: str = obj[ID]
    id_target_prio: int = _get_target_priority(target, id)
    property: Optional[str] = id if id_target_prio else None
    prop_prio: int = id_target_prio

    # TODO: OK to add identity-mapping?
    if property:
        _add_rule(target_map, id, property, id_target_prio)

    candidates: Candidates = _collect_candidates(obj, rels)

    baseprops: List[Tuple[int, str]] = []
    if id_target_prio:
        baseprops.append((id_target_prio, id))

    candidate_prop: Optional[str] = None

    while candidates:
        crel, candidate = candidates.pop(0)
        if ID not in candidate:
            continue
        candidate_id: str = candidate[ID]
        candidate = cast(Dict, vocab_index.get(candidate_id, candidate))

        target_prio: int = _get_target_priority(target, candidate_id)
        if not id_target_prio and target_prio:
            baseprops.append((target_prio, candidate_id))
            _add_rule(target_map, id, candidate_id, target_prio)
            candidate_prop = candidate_id
            prop_prio = target_prio
            #break
        elif crel in SYMMETRIC and not target_prio and id_target_prio:
            _add_rule(target_map, candidate_id, id, id_target_prio)
            candidate_prop = candidate_id
            prop_prio = target_prio
            #break
        else:
            _extend_candidates(candidates, candidate, rels)

    _process_property_chain(obj, target, target_map, candidate_prop, baseprops)


def _process_property_chain(obj: Dict,
                             target: Dict[str, object],
                             target_map: Dict,
                             candidate_prop: Optional[str],
                             baseprops: List[Tuple[int, str]]) -> bool:
    if OWL_propertyChainAxiom not in obj:
        return False

    prop_chain: List[Dict] = cast(List[Dict], obj[OWL_propertyChainAxiom])
    source_property: Optional[str] = None

    lst: List[Dict] = prop_chain[0][LIST]
    lead: Dict = lst[0]
    # TODO: assert(len(lst) == 2) or use rest as path to value...
    if lead:
        source_property = lead.get(ID)

    value_from: str = lst[1][ID]
    rtype: Optional[str] = None

    # TODO: don't rely solely on anonymous subPropertyOf
    if source_property is None or source_property.startswith('_:'):
        try:
            ranges: List[Dict] = lead[RDFS_range]
            rtype = ranges[0][ID]
        except:
            pass
        superprops: List[Dict] = lead[RDFS_subPropertyOf]
        source_property = superprops[0][ID]

    match: Optional[Dict] = {TYPE: rtype} if rtype else None
    # TODO: also match OWL_Restriction using OWL_onProperty PLUS
    # OWL_hasValue OR OWL_allValuesFrom

    if source_property:
        if (source_property != candidate_prop and
            not _get_target_priority(target, source_property)):
            for prio, baseprop in baseprops:
                rule: Dict = _rule_from(baseprop, None, value_from, match)
                _add_rule(target_map, source_property, rule, prio)

            return True

    # TODO: also check vocab_index[source_property][REVERSE][RDFS_subPropertyOf]

    return False


def _collect_candidates(obj: Dict, rels: List[str]) -> Candidates:
    candidates: Candidates = []
    for rel in rels:
        refs: object = obj.get(rel)
        if isinstance(refs, List):
            candidates += cast(Candidates, [(rel, ref) for ref in refs])
    return candidates


def _extend_candidates(candidates: Candidates, candidate: Dict, rels: List[str]):
    for rel in rels:
        superrefs: Optional[List[Dict]] = cast(Optional[List[Dict]], candidate.get(rel))
        if superrefs is not None:
            for sup in superrefs:
                if ID in sup:
                    candidates.append((None, sup))


def _process_reified_forms(obj: Dict, vocab_index: Dict, target_map: Dict[str, object]):
    prop = _trace_inverse_of_subject(obj, vocab_index)

    if prop is not None:
        ranges: List[Dict] = []

        prop_id: str = obj[ID]

        if RDFS_range in obj:
            ranges += as_list(obj[RDFS_range])
        if prop is not obj and RDFS_range in prop:
            ranges += as_list(prop[RDFS_range])

        property_from: Optional[str] = None
        value_from: Optional[str] = None

        for range in ranges:
            range_node: Optional[Dict] = vocab_index.get(range[ID])
            if range_node is None:
                continue
            reverses: Optional[Dict] = cast(Dict, range_node.get(REVERSE))
            in_domain_of: List[Dict] = cast(List[Dict], reverses.get(RDFS_domain, [])) if reverses is not None else []
            for domain_prop in in_domain_of:
                if leads_to(domain_prop, vocab_index, RDFS_subPropertyOf, RDF_predicate):
                    property_from = domain_prop[ID]
                elif leads_to(domain_prop, vocab_index, RDFS_subPropertyOf, RDF_object):
                    value_from = domain_prop[ID]

        if property_from and value_from and isinstance(prop_id, str):
            _add_rule(target_map, prop_id, _rule_from(None, property_from, value_from, None))


def _trace_inverse_of_subject(obj: Dict, vocab_index: Dict) -> Optional[Dict]:
    invs = cast(Optional[List[Dict]], obj.get(OWL_inverseOf))

    if invs is not None:
        for p in invs:
            if leads_to(p, vocab_index, RDFS_subPropertyOf, RDF_subject):
                return p

    supers = cast(Optional[List[Dict]], obj.get(RDFS_subPropertyOf))
    if supers is None:
        return None

    for supref in supers:
        sup = cast(Dict, vocab_index.get(supref[ID], supref) if ID in supref else supref)
        if _trace_inverse_of_subject(sup, vocab_index):
            return sup

    return None


def _add_rule(target_map: Dict[str, object],
              source_id: str,
              rule: Union[str, List, Dict],
              priority=0):
    if source_id == rule:
        return

    rule_priority: object = [(priority, it) for it in rule] if isinstance(rule, List) else [(priority, rule)]

    rules: List = cast(List, target_map.get(source_id))
    if rules is None:
        target_map[source_id] = rule_priority
    else:
        if not isinstance(rules, List):
            rules = [rules]
            target_map[source_id] = rules
        if isinstance(rule_priority, List):
            rules += rule_priority
        else:
            rules.append(rule_priority)


def _rule_from(property: Optional[str],
               property_from: Optional[str],
               value_from: Optional[str],
               match: Optional[Dict[str, str]]) -> Dict:
    return {
        'property': property,
        'propertyFrom': property_from,
        'valueFrom': value_from,
        'match': match
    }


def _get_target_priority(target: Dict[str, object], id: str) -> int:
    top_prio: int = len(target)

    prio: int = top_prio * 3
    for v in target.values():
        if id == v:
            return prio
        prio -= 1

    if VOCAB in target and id.startswith(cast(str, target[VOCAB])):
        return top_prio * 2

    prio = top_prio
    for v in target.values():
        if isinstance(v, str) and id.startswith(cast(str, v)):
            return prio
        prio -= 1

    return 0


def leads_to(s: Dict, vocab_index: Dict, rel: str, o) -> bool:
    if s.get(ID) == o:
        return True

    data: Optional[Dict] = cast(Dict, vocab_index.get(s[ID], s) if ID in s else s)
    xs: List[Dict] = cast(List, data.get(rel, [])) if data is not None else []

    for x in xs:
        if x.get(ID) == o or leads_to(x, vocab_index, rel, o):
            return True

    return False



import json
import sys
import os
from ..jsonld.base import CONTEXT
from ..jsonld.expansion import expand
from ..jsonld.compaction import compact
from .mapmaker import make_target_map
from .mapper import map_to


args = sys.argv[1:]
if len(args) < 2:
    print('Usage: python -m trld.tvm VOCAB TARGET [INFILE]', file=sys.stderr)
    sys.exit(0)

vocabfile = args.pop(0)
targetref = args.pop(0)
infile = args.pop(0) if args else None
drop_unmapped = '--drop' in args

if os.path.isfile(targetref):
    with open(targetref) as f:
        target = json.load(f)
else:
    target = {"@context": {"@vocab": targetref}}

with open(vocabfile) as f:
    vocab = json.load(f)
    vocab = expand(vocab, vocabfile)

target_map = make_target_map(vocab, target)

if not infile:
    print(json.dumps(target_map, indent=2))
else:
    with open(infile) as f:
        indata = json.load(f)
        indata = expand(indata, infile)

    outdata = map_to(target_map, indata, drop_unmapped)
    outdata = compact(target, outdata) # type: ignore
    if isinstance(outdata, dict):
        outdata[CONTEXT] = target[CONTEXT]

    print(json.dumps(outdata, indent=2))



from typing import Dict, List, Tuple, Optional, cast

from ..common import Input, Output, load_json
from .base import *
from . import context
from .expansion import expand
from .compaction import compact
from .flattening import flatten
from .rdf import RdfDataset, to_jsonld, to_rdf_dataset
#from ..nq import parser as nq_parser, serializer as nq_serializer
from ..nq.parser import load
from ..nq.serializer import serialize


TESTS_URL: str = 'https://w3c.github.io/json-ld-api/tests'


class TestCase:

    testtype: List[str]
    testid: str
    name: str
    options: Dict[str, object]
    indoc_path: str
    base_uri: str

    compact_arrays: bool
    expected_error: Optional[str]
    expectdoc_path: Optional[str]
    expand_context: Optional[str]
    context_path: Optional[str]
    compact_context: object

    def __init__(self, suitedir: str, tc_data: Dict):
        self.testtype = tc_data[TYPE]
        self.testid = cast(str, tc_data[ID]).replace('#t', '')
        self.name = tc_data['name']

        self.options = cast(Dict, tc_data.get(u'option', {}))

        self.indoc_path = f"{suitedir}/{tc_data['input']}"

        base_uri: str = self.indoc_path.replace(suitedir, TESTS_URL)
        self.base_uri = cast(str, self.options.get('base', base_uri))

        self.compact_arrays = cast(bool, self.options.get('compactArrays', True))

        self.expected_error = tc_data.get('expectErrorCode')

        self.expectdoc_path = None
        expectdoc_path: Optional[str] = tc_data.get('expect')
        if expectdoc_path:
            self.expectdoc_path = f'{suitedir}/{expectdoc_path}'

        self.expand_context = None
        expand_context: Optional[str] = cast(Optional[str], self.options.get('expandContext'))
        if expand_context:
            self.expand_context = f'{suitedir}/{expand_context}'

        self.context_path = None
        self.compact_context = None
        context_path: str = cast(str, tc_data.get('context'))
        if context_path:
            self.context_path = f'{suitedir}/{context_path}'
            self.compact_context = load_json(str(self.context_path))

    def run(self) -> Tuple[object, object]:
        out_data: object = None
        if 'jld:FromRDFTest' in self.testtype:
            out_data = self.run_from_rdf_test()

        elif 'jld:ToRDFTest' in self.testtype:
            out_data = self.run_to_rdf_test()

        elif 'jld:ExpandTest' in self.testtype \
                or 'jld:CompactTest' in self.testtype \
                or 'jld:FlattenTest' in self.testtype:
            in_data: object = load_json(self.indoc_path)
            out_data = expand(cast(JsonObject, in_data), self.base_uri,
                    expand_context=self.expand_context, ordered=True)

        if 'jld:FlattenTest' in self.testtype:
            out_data = flatten(cast(JsonObject, out_data), ordered=True)

        if self.compact_context:
            out_data = compact(self.compact_context, cast(JsonObject, out_data),
                    self.base_uri,
                    compact_arrays=self.compact_arrays, ordered=True)

        return out_data, self._load_expect_data()

    def run_from_rdf_test(self) -> object:
        in_data: RdfDataset = RdfDataset()
        with Input(self.indoc_path) as inp:
            load(in_data, inp)
        ordered: bool = True
        rdf_direction: Optional[str] = cast(Optional[str], self.options.get('rdfDirection'))
        use_native_types: bool = cast(bool, self.options.get('useNativeTypes', False))
        use_rdf_type: bool = cast(bool, self.options.get('useRdfType', False))
        return to_jsonld(in_data, ordered,
                rdf_direction, use_native_types, use_rdf_type)

    def run_to_rdf_test(self) -> object:
        out: Output = Output()
        in_data: JsonObject = cast(JsonObject, load_json(self.indoc_path))
        in_data = expand(in_data, self.base_uri,
                expand_context=self.expand_context, ordered=True)
        #in_data = flatten(cast(JsonObject, in_data), ordered=True)
        rdf_direction: Optional[str] = cast(Optional[str], self.options.get('rdfDirection'))
        dataset: RdfDataset = to_rdf_dataset(in_data, rdf_direction)
        serialize(dataset, out)
        return out.get_value()

    def _load_expect_data(self) -> Optional[object]:
        if self.expectdoc_path:
            assert 'jld:PositiveEvaluationTest' in self.testtype
            if self.expectdoc_path.endswith('.nq'):
                with Input(self.expectdoc_path) as inp:
                    return inp.read()
            else:
                expect_data: object = load_json(self.expectdoc_path)
                if self.context_path and isinstance(expect_data, Dict):
                    expect_data.pop(CONTEXT, None)
                return expect_data
        elif 'jld:PositiveSyntaxTest' in self.testtype:
            pass # NOTE: only input; just test conversion!
        else:
            assert 'jld:NegativeEvaluationTest' in self.testtype
        return None



import json
from ..jsonld.base import CONTEXT, GRAPH, ID
from ..jsonld.expansion import expand
from ..jsonld.compaction import compact
from ..jsonld.extras.index import make_index
from .mapmaker import make_target_map, leads_to
from .mapper import map_to


DEBUG = False

null = None


context = {
    "@context": {
        "ex": "http://example.org/ns#",
        "rdf": "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
        "rdfs": "http://www.w3.org/2000/01/rdf-schema#",
        "xsd": "http://www.w3.org/2001/XMLSchema#",
        "owl": "http://www.w3.org/2002/07/owl#",
        "skos": "http://www.w3.org/2004/02/skos/core#",
        "bibo": "http://purl.org/ontology/bibo/",
        "dc": "http://purl.org/dc/terms/",
        "foaf": "http://xmlns.com/foaf/0.1/",
        "schema": "http://schema.org/",
        "bf": "http://id.loc.gov/ontologies/bibframe/",
        "lcrel": "http://id.loc.gov/vocabulary/relators/",
        "edtf": "http://id.loc.gov/datatypes/edtf/"
    }
}


def test_make_target_map():
    vocab = {
        "@graph": [
            {
                "@id": "foaf:Document",
                "rdfs:subClassOf": {"@id": "rdfs:Resource"}
            },
            {
                "@id": "dc:title",
                "rdfs:subPropertyOf": {"@id": "rdfs:label"}
            }
        ]
    }
    target = {"@vocab": "http://www.w3.org/2000/01/rdf-schema#"}
    expect = {
        "http://xmlns.com/foaf/0.1/Document": [
            "http://www.w3.org/2000/01/rdf-schema#Resource"
        ],
        "http://purl.org/dc/terms/title": [
            "http://www.w3.org/2000/01/rdf-schema#label"
        ],
        "http://www.w3.org/2000/01/rdf-schema#Resource": "http://www.w3.org/2000/01/rdf-schema#Resource",
        "http://www.w3.org/2000/01/rdf-schema#label": "http://www.w3.org/2000/01/rdf-schema#label",
    }
    vocab = expand(dict(context, **vocab), "")
    target_map = make_target_map(vocab, {CONTEXT: target})
    assert_json_equals(target_map, expect)


def test_cope_with_circular_references():
    target = {"@vocab": "http://www.w3.org/2000/01/rdf-schema#"}

    assuming = {
        "@graph": [
            {
                "@id": "owl:Thing",
                "rdfs:subClassOf": {"@id": "skos:Concept"}
            },
            {
                "@id": "skos:Concept",
                "rdfs:subClassOf": {"@id": "owl:Thing"}
            }
        ]
    }

    _to_target_map(target, assuming)

    print_ok()


def test_reducing_foaf_to_rdfs():
    given = {
        "@id": "",
        "@type": "foaf:Document",
        "dc:title": "A"
    }

    expect = {
        "@id": "",
        "@type": "rdfs:Resource",
        "rdfs:label": "A"
    }

    target = {"@vocab": "http://www.w3.org/2000/01/rdf-schema#"}

    assuming = {
        "@graph": [
            {
                "@id": "foaf:Document",
                "rdfs:subClassOf": {"@id": "rdfs:Resource"}
            },
            {
                "@id": "dc:title",
                "rdfs:subPropertyOf": {"@id": "rdfs:label"}
            }
        ]
    }

    check(**locals())


def test_remap_datatypes():
    given = {
        "@id": "",
        "dc:issued": {"@type": "2021", "@type": "xsd:year"}
    }

    expect = {
        "@id": "",
        "dc:issued": {"@type": "2021", "@type": "edtf:EDTF-level0"}
    }

    target = {
        "dc": "http://purl.org/dc/terms/",
        "edtf": "http://id.loc.gov/datatypes/edtf/"
    }

    assuming = {
        "@graph": [
            {
                "@id": "xsd:year",
                "rdfs:subClassOf": [
                    {"@id": "edtf:EDTF-level0"}
                ]
            }
        ]
    }

    check(**locals())


def test_structured_values_and_shorthand_properties():
    given = {
        "@id": "/instance/a",
        "bf:identifiedBy": {
            "@type": "bf:Isbn",
            "rdf:value": "12-3456-789-0"
        }
    }

    target = {"@vocab": "http://schema.org/"}

    expect = {
        "@id": "/instance/a",
        "schema:isbn": "12-3456-789-0"
    }

    assuming = {
        "@graph": [
            {
                "@id": "schema:isbn",
                "owl:propertyChainAxiom": {
                    "@list": [
                        {
                            "rdfs:subPropertyOf": {"@id": "bf:identifiedBy"},
                            "rdfs:range": {"@id": "bf:Isbn"}
                        },
                        {"@id": "rdf:value"}
                    ]
                }
            }
        ]
    }

    check(**locals())


def test_qualified_relations_as_reifications():
    given = {
        "@id": "/work",
        "bf:contribution": {
            "bf:agent": {"@id": "/person/a"},
            "bf:role": {"@id": "lcrel:ill"}
        }
    }

    target = {"@vocab": "http://purl.org/dc/terms/"}

    expect = {
        "@id": "/work",
        "dc:contributor": {"@id": "/person/a"}
    }

    assuming = {
        "@graph": [
            {
                "@id": "bf:Contribution",
                "rdfs:subClassOf": [
                    {"@id": "rdf:Statement"},
                    #{
                    #    "owl:onProperty": {"@id": "rdf:predicate"},
                    #    "owl:hasValue": {"@id": "dc:contributor"}
                    #}
                ]
            },
            {
                "@id": "bf:contribution",
                "rdfs:range": {"@id": "bf:Contribution"},
                "rdfs:subPropertyOf": {
                    "owl:inverseOf": {"@id": "rdf:subject"}
                }
            },
            {
                "@id": "bf:role",
                "rdfs:domain": {"@id": "bf:Contribution"},
                "rdfs:subPropertyOf": {"@id": "rdf:predicate"}
            },
            {
                "@id": "bf:agent",
                "rdfs:domain": {"@id": "bf:Contribution"},
                "rdfs:subPropertyOf": {"@id": "rdf:object"}
            },
            {
                "@id": "lcrel:ill",
                "rdfs:subPropertyOf": {"@id": "dc:contributor"}
            }
        ]
    }

    check(**locals())


def test_inferred_qualified_relations_as_reifications():
    given = {
        "@id": "/work",
        "bf:contribution": {
            "bf:agent": {"@id": "/person/a"},
            "bf:role": {"@id": "lcrel:ill"}
        }
    }

    target = {"@vocab": "http://purl.org/dc/terms/"}

    expect = {
        "@id": "/work",
        "dc:contributor": {"@id": "/person/a"}
    }

    assuming = {
        "@graph": [
            {
                "@id": "ex:QualifiedEvent",
                "rdfs:subClassOf": [
                    {"@id": "rdf:Statement"}
                ]
            },
            {
                "@id": "ex:subject",
                "rdfs:subPropertyOf": {"@id": "rdf:subject"}
            },
            {
                "@id": "ex:role",
                "rdfs:subPropertyOf": {"@id": "rdf:predicate"}
            },
            {
                "@id": "ex:object",
                "rdfs:subPropertyOf": {"@id": "rdf:object"}
            },

            {
                "@id": "ex:hasStatement",
                "rdfs:subPropertyOf": {
                    "owl:inverseOf": {"@id": "ex:subject"}
                }
            },

            {
                "@id": "bf:Contribution",
                "rdfs:subClassOf": [
                    {"@id": "ex:QualifiedEvent"}
                ]
            },
            {
                "@id": "bf:contribution",
                "rdfs:range": {"@id": "bf:Contribution"},
                "rdfs:subPropertyOf": {"@id": "ex:hasStatement"}
            },
            {
                "@id": "bf:role",
                "rdfs:domain": {"@id": "bf:Contribution"},
                "rdfs:subPropertyOf": {"@id": "ex:role"}
            },
            {
                "@id": "bf:agent",
                "rdfs:domain": {"@id": "bf:Contribution"},
                "rdfs:subPropertyOf": {"@id": "ex:object"}
            },
            {
                "@id": "lcrel:ill",
                "rdfs:subPropertyOf": {"@id": "dc:contributor"}
            }
        ]
    }

    check(**locals())


def test_property_chains_for_events():
    given = {
        "@id": "/instance/a",
        "bf:provisionActivity": {
            "@type": "bf:Publication",
            "bf:agent": {"@id": "/org/a"},
            "bf:date": "2017"
        }
    }

    target = {"@vocab": "http://schema.org/"}

    expect = {
        "@id": "/instance/a",
        "schema:publisher": {"@id": "/org/a"},
        "schema:datePublished": "2017"
    }

    assuming = {
        "@graph": [
            {
                "@id": "schema:datePublished",
                "rdfs:subPropertyOf": {"@id": "dcterms:issued"},
                "owl:propertyChainAxiom": {
                    "@list": [
                        {
                            "rdfs:subPropertyOf": {"@id": "bf:provisionActivity"},
                            "rdfs:range": {"@id": "bf:Publication"}
                        },
                        {"@id": "bf:date"}
                    ]
                }
            },
            {
                "@id": "dcterms:publisher",
                "owl:equivalentProperty": {"@id": "schema:publisher"},
                "owl:propertyChainAxiom": {
                    "@list": [
                        {
                            "rdfs:subPropertyOf": {"@id": "bf:provisionActivity"},
                            "rdfs:range": {"@id": "bf:Publication"}
                        },
                        {"@id": "bf:agent"}
                    ]
                }
            }
        ]
    }

    check(**locals())


def test_add_to_existing_key():
    given = {
        "@id": "",
        "dc:title": "Title",
        "foaf:name": "Name",
        "rdfs:label": "Label"
    }

    expect = {
        "@id": "",
        "rdfs:label": ["Title", "Name", "Label"]
    }

    target = {"@vocab": "http://www.w3.org/2000/01/rdf-schema#"}

    assuming = {
        "@graph": [
            {
                "@id": "foaf:name",
                "rdfs:subPropertyOf": {"@id": "rdfs:label"}
            },
            {
                "@id": "dc:title",
                "rdfs:subPropertyOf": {"@id": "rdfs:label"}
            }
        ]
    }

    check(**locals())


def test_only_add_most_specific():
    given = {
        "@id": "",
        "bf:identifiedBy": {
            "rdf:value": "0000000000"
        }
    }

    expect = {
        "@id": "",
        "bibo:identifier": "0000000000"
    }

    target = [
        {"bibo": "http://purl.org/ontology/bibo/"},
        {"dc": "http://purl.org/dc/terms/"},
        {"rdfs": "http://www.w3.org/2000/01/rdf-schema#"}
    ]

    assuming = {
        "@graph": [
            {
                "@id": "bibo:identifier",
                "rdfs:subPropertyOf": [{"@id": "dc:identifier"}, {"@id": "rdfs:label"}],
                "owl:propertyChainAxiom": {
                    "@list": [
                        {"@id": "bf:identifiedBy"},
                        {"@id": "rdf:value"}
                    ]
                }
            }
        ]
    }

    check(**locals())


def test_sort_target_rules():
    vocab = {
        "@graph": [
            {
                "@id": "schema:identifier",
                #"rdfs:subPropertyOf": {"@id": "dc:identifier"},
                "owl:propertyChainAxiom": {
                    "@list": [
                        {"@id": "bf:identifiedBy"},
                        {"@id": "rdf:value"}
                    ]
                }
            },
            {
                "@id": "schema:isbn",
                "owl:propertyChainAxiom": {
                    "@list": [
                        {
                            "rdfs:subPropertyOf": {"@id": "bf:identifiedBy"},
                            "rdfs:range": {"@id": "bf:Isbn"}
                        },
                        {"@id": "rdf:value"}
                    ]
                }
            }
        ]
    }
    target = [
        {"@vocab": "http://schema.org/"},
        #{"@vocab": "http://purl.org/dc/terms/"}
    ]
    expect = {
        "http://id.loc.gov/ontologies/bibframe/identifiedBy": [
            {
                "match": {
                    "@type": "http://id.loc.gov/ontologies/bibframe/Isbn"
                },
                "property": "http://schema.org/isbn",
                "propertyFrom": None,
                "valueFrom": "http://www.w3.org/1999/02/22-rdf-syntax-ns#value"
            },
            {
                "match": null,
                "property": "http://schema.org/identifier",
                "propertyFrom": null,
                "valueFrom": "http://www.w3.org/1999/02/22-rdf-syntax-ns#value"
            }
        ]
    }
    vocab = expand(dict(context, **vocab), "")
    target_map = make_target_map(vocab, {CONTEXT: target})
    assert_json_equals(target_map, expect)



def test_keep_target_terms():
    given = {
        "@id": "",
        "@type": "rdfs:Resource",
        "rdfs:label": "A"
    }

    expect = given

    target = {"@vocab": "http://www.w3.org/2000/01/rdf-schema#"}

    assuming = {
        "@graph": [
            {
                "@id": "foaf:Document",
                "rdfs:subClassOf": {"@id": "rdfs:Resource"}
            },
            {
                "@id": "dc:title",
                "rdfs:subPropertyOf": {"@id": "rdfs:label"}
            }
        ]
    }

    check(**locals())


def test_leads_to():
    r0 = {ID: "urn:x-test:0"}
    r1 = {ID: "urn:x-test:1"}
    r2 = {ID: "urn:x-test:2"}
    rel = "urn:x-test:rel"
    graph = [
        r0,
        dict(r1, **{rel: [r0]}),
        dict(r2, **{rel: [r1]})
    ]
    index = make_index(graph)
    assert leads_to(r0, index, rel, r0[ID])
    assert leads_to(r1, index, rel, r0[ID])
    assert leads_to(r2, index, rel, r0[ID])
    print_ok()


def check(given, target, expect, assuming):
    target_map = _to_target_map(target, assuming)
    if DEBUG:
        print(_jsonstr(target_map))
    indata = expand(dict(context, **given), "")
    outdata = map_to(target_map, indata)
    outdata = compact(context, outdata)
    assert_json_equals(outdata, expect)


def assert_json_equals(given, expected):
    gvn = _jsonstr(given)
    exp = _jsonstr(expected)
    assert_equals(gvn, exp)


def assert_equals(given: str, expected: str):
    assert given == expected, f'Expected:\n{expected}\nGot:\n{given}'
    print_ok()


def print_ok():
    print('OK', end='')


def _to_target_map(target, assuming) -> dict:
    vocab = expand(dict(context, **assuming), "")
    return make_target_map(expand(vocab, ""), {CONTEXT: target})


def _jsonstr(data) -> str:
    return json.dumps(data, indent=2, sort_keys=True)


if __name__ == '__main__':
    from sys import argv

    DEBUG = '-d' in argv

    for k, v in dict(globals()).items():
        if k.startswith('test_') and hasattr(v, '__call__'):
            print(f'Running {k}:', end=' ')
            try:
                v()
            except AssertionError as e:
                print(str(e))
            else:
                print()



from typing import Callable, Dict, List, NamedTuple, Optional, Union, cast
import re

from ..common import Output, uuid4
from ..jsonld.base import (BASE, CONTAINER, CONTEXT, GRAPH, ID, INDEX,
                           LANGUAGE, LIST, PREFIX, PREFIX_DELIMS, REVERSE,
                           TYPE, VALUE, VOCAB)

StrObject = Dict[str, object]
StrOrObject = Union[str, StrObject]

ANNOTATION: str = '@annotation'

WORD_START = re.compile(r'^\w*$')
PNAME_LOCAL_ESC = re.compile(r"([~!$&'()*+,;=/?#@%]|^[.-]|[.-]$)")


class Settings(NamedTuple):
    turtle_only: bool = False
    turtle_drop_named: bool = False
    drop_rdfstar: bool = False
    indent_chars: str = '  '
    use_graph_keyword: bool = True
    upcase_keywords: bool = False
    # space_before_semicolon: bool = True
    predicate_repeat_new_line: bool = True
    # bracket_start_new_line: bool = False
    bracket_end_new_line: bool = False
    # list_item_new_line: bool = True
    prologue_end_line: int = 1
    # separator_lines: int = 1


class KeyAliases(NamedTuple):
    id: str = ID
    value: str = VALUE
    type: str = TYPE
    lang: str = LANGUAGE
    graph: str = GRAPH
    list_: str = LIST
    reverse: str = REVERSE
    index_: str = INDEX
    annotation: str = ANNOTATION


def serialize(
    data: StrObject,
    out: Output,
    context: Optional[Dict] = None,
    base_iri: Optional[str] = None,
    settings: Settings = None,
):
    settings = settings if settings is not None else Settings()
    state = SerializerState(out, settings, context, base_iri)
    state.serialize(data)


def serialize_turtle(
    data: StrObject,
    out: Output,
    context: Optional[Dict] = None,
    base_iri: Optional[str] = None,
    union: bool = False,
):
    settings = Settings(turtle_only=True, turtle_drop_named=not union)
    serialize(data, out, context, base_iri, settings)


class SerializerState:
    settings: Settings
    out: Output
    parent: Optional['SerializerState']
    context: StrObject
    base_iri: Optional[str]
    prefixes: Dict[str, str]
    aliases: KeyAliases
    bnode_skolem_base: Optional[str] = None
    prefix_keyword: str
    base_keyword: str
    graph_keyword: Optional[str] = None
    unique_bnode_suffix: str
    bnode_counter: int

    def __init__(
            self,
            out: Output,
            settings: Settings,
            context: Optional[object],
            base_iri: Optional[str] = None,
            parent: 'SerializerState' = None,
    ):
        self.out = out if out is not None else cast(SerializerState, parent).out

        self.base_iri = base_iri
        self.parent = parent
        self.settings = parent.settings if parent is not None else settings
        self.prefix_keyword = self._kw('prefix')
        self.base_keyword = self._kw('base')
        if self.settings.use_graph_keyword:
            self.graph_keyword = self._kw('graph')
        self.unique_bnode_suffix = ''
        self.bnode_counter = 0

        self.aliases = KeyAliases()
        self.context = {}
        self.prefixes = {}
        self.init_context(context)

    def _kw(self, s: str) -> str:
        return s.upper() if self.settings.upcase_keywords else s

    def init_context(self, ctx: Optional[object]):
        # TODO: Only supports limited "Turtle-like" subset. Parse using jsonld?
        merged: StrObject = {}

        if len(self.context) > 0:
            merged.update(self.context)

        if isinstance(ctx, List):
            for item in ctx:
                merged.update(cast(StrObject, item))
        elif isinstance(ctx, Dict):
                merged.update(ctx)

        if len(merged) > 0:
            self.context = merged
            self.prefixes = collect_prefixes(merged)

    def serialize(self, data: Union[List, Dict]):
        if isinstance(data, Dict):
            self.init_context(data.get(CONTEXT))
            self.prelude(self.prefixes)
            self.write_object(data)
        else:
            assert isinstance(data, List)
            self.prelude(self.prefixes)
            for item in data:
                self.write_object(item)

    def prelude(self, prefixes: Dict[str, str]):
        for k, v in prefixes.items():
            if k == BASE:
                self.write_base(v)
            else:
                self.writeln(f'{self.prefix_keyword} {k}: <{v}>')
        if self.base_iri:
            self.write_base(self.base_iri)
        if self.settings.prologue_end_line > 1:
            self.writeln()

    def write_base(self, iri: str):
        self.writeln(f'{self.base_keyword} <{iri}>')

    def is_list_container(self, term: str) -> bool:
        return self._is_container(term, LIST)

    def is_lang_container(self, term: str) -> bool:
        return self._is_container(term, LANGUAGE)

    def _is_container(self, term: str, kind: str) -> bool:
        if self.context is not None:
            termdef: object = self.context.get(term)
            if isinstance(termdef, Dict):
                return termdef.get(CONTAINER) == kind

        return False

    def write_graph(self, iri: Optional[str], graph: object, depth: int = 0):
        if iri is not None and self.settings.turtle_drop_named:
            return

        in_graph_block: bool = iri is not None or depth > 0

        if not self.settings.turtle_only:
            if iri is None:
                if depth > 0:
                    self.writeln()
                    self.writeln("{")
            else:
                self.writeln()
                if self.graph_keyword:
                        self.write(f'{self.graph_keyword} ')
                self.writeln(f'{self.ref_repr(iri)} {{')

        for node in as_list(graph):
            via: Optional[str] = self.aliases.graph if in_graph_block else None
            self.write_object(cast(StrObject, node), depth, via)

        if not self.settings.turtle_only:
            if in_graph_block:
                self.writeln()
                self.writeln("}")

    def write_object(
            self,
            obj: object,
            depth: int = 0,
            via_key: Optional[str] = None,
    ) -> List[StrObject]:
        if depth > 0 and isinstance(obj, Dict) and CONTEXT in obj:
            # TODO: use SerializerState(node[CONTEXT], None, None, state)
            raise Exception('Nested context not supported yet')

        if via_key and self.is_lang_container(via_key) and isinstance(obj, Dict):
            first = True
            for lang, value in cast(StrObject, obj).items():
                if not first:
                    self.write(' , ')
                self.to_literal(
                 { self.aliases.value: value, self.aliases.lang: lang },
                    via_key)
                first = False
            return []

        if not isinstance(obj, Dict) or self.aliases.value in obj:
            self.to_literal(cast(object, obj), via_key)
            return []

        explicit_list: bool = self.aliases.list_ in obj

        if via_key and self.is_list_container(via_key):
            obj = { self.aliases.list_: obj }

        s = cast(Optional[str], obj.get(self.aliases.id))

        is_list: bool = self.aliases.list_ in obj
        started_list: bool = is_list

        is_bracketed: bool = is_list or via_key == self.aliases.annotation

        if self.aliases.graph in obj:
            if s is not None and self.settings.turtle_drop_named:
                return []

            if CONTEXT in obj and depth > 0:
                self.prelude(collect_prefixes(obj[CONTEXT]))
            self.write_graph(s, obj[self.aliases.graph], depth)
            return []

        if explicit_list:
            self.write('( ')

        in_graph: bool = via_key == self.aliases.graph and not self.settings.turtle_only
        in_graph_add: int = 1 if in_graph else 0

        if (s is not None or depth == 0) and self.has_keys(obj, 2):
            if s is None:
                self.write("[]")
            else:
                if depth == 0:
                    self.writeln()
                if in_graph_add > 0:
                    self.write(self.get_indent(0))
                self.write(self.ref_repr(s))
        elif depth > 0:
            if not is_bracketed:
                depth += 1
                self.write("[")
        else:
            return []

        indent = self.get_indent(depth + in_graph_add)

        nested_depth: int = depth + 1 + in_graph_add

        top_objects: List[StrObject] = []

        first = True
        ended_list = False

        for key, vo in cast(StrObject, obj).items():

            index_key: Optional[str] = self.index_key_for(key)
            if index_key is not None:
                key = index_key
                vo = list(vo.values()) if isinstance(vo, Dict) else vo

            term = self.term_for(key)

            rev_key: Optional[str] = self.rev_key_for(key) if term is None else None
            if term is None and rev_key is None:
                continue

            if term == self.aliases.id or term == CONTEXT:
                continue

            if term == self.aliases.index_:
                continue

            if term == self.aliases.annotation:
                continue

            vs: List = vo if isinstance(vo, List) else [vo] if vo is not None else []
            vs = cast(List, [x for x in vs if x is not None])

            if len(vs) == 0: # TODO: and not @list
                continue

            in_list: bool = is_list or self.is_list_container(key)

            rev_container: Optional[StrObject] = None
            if term == self.aliases.reverse:
                rev_container = cast(Optional[StrObject], obj[key])
            elif rev_key:
                rev_container = { rev_key: obj[key] }

            if rev_container:
                for revkey, rvo in rev_container.items():
                    vs = rvo if isinstance(rvo, List) else [rvo] if rvo is not None else []
                    for x in vs:
                        top_objects.append(
                            self.make_top_object(s, revkey, cast(StrObject, x))
                        )
            else:
                use_indent = indent
                if first:
                    use_indent = ' '
                    first = False
                else:
                    if started_list and not in_list and not ended_list:
                        ended_list = True
                        self.write(" )")
                    self.writeln(" ;")

                assert isinstance(term, str)

                if term == self.aliases.type:
                    term = "a"

                if term != LIST:
                    term = self.to_valid_term(term)
                    self.write(use_indent + term + " ")

                for i in range(len(vs)):
                    v: object = vs[i]

                    if in_list:
                        if not started_list:
                            self.write("(")
                            started_list = True
                        self.write(" ")
                    elif i > 0:
                        if self.settings.predicate_repeat_new_line:
                            self.writeln(' ,')
                            self.write(self.get_indent(nested_depth))
                        else:
                            self.write(' , ')

                    if self.bnode_skolem_base and isinstance(v, Dict) and self.aliases.id not in v:
                        s = self.gen_skolem_id()
                        v[self.aliases.id] = s

                    if term == "a":
                        t = self.repr_type(cast(StrOrObject, v))
                        self.write(t)
                    elif v and isinstance(v, Dict) and self.aliases.id in v:
                        top_objects.append(v)
                        self.write(self.ref_repr(v[self.aliases.id]))
                    elif v is not None:
                        objects = self.write_object(v, nested_depth, key)
                        for it in objects:
                            top_objects.append(it)

                    self.write_annotation(v, depth)

        if explicit_list or (not is_list and started_list) and not ended_list:
            self.write(" )")

        if depth == 0:
            if not first:
                self.writeln(" .")
            #self.writeln()
            for it in top_objects:
                self.write_object(it, depth, via_key)
            return []
        else:
            indent = self.get_indent(nested_depth - (1 + in_graph_add))
            if self.settings.bracket_end_new_line:
                self.writeln()
                self.write(indent)
            else:
                self.write(' ')
            if not is_bracketed:
                # NOTE: hack for e.g. BlazeGraph
                #if not self.has_keys(obj) and self.settings.mark_empty_bnode:
                #    self.writeln(f'a {self.settings.empty_marker}')
                #    self.write(indent)
                self.write("]")
            return top_objects

    def write_annotation(self, v: object, depth: int):
        if self.settings.drop_rdfstar:
            return

        if isinstance(v, Dict) and self.aliases.annotation in v:
            annotation: StrObject = v[self.aliases.annotation]
            if annotation is not None:
                self.write(' {|')
                self.write_object(annotation, depth + 2, self.aliases.annotation)
                self.write('|}')

    def to_literal(
        self,
        obj: object,
        via_key: Optional[str] = None,
        write: Callable = None, # [[str], None]
    ):
        if write is None:
            write = lambda s: self._write(s)
        value = obj
        lang: Optional[object] = self.context.get(LANGUAGE)
        datatype: Optional[str] = None
        if isinstance(obj, Dict):
            value = obj.get(self.aliases.value)
            datatype = cast(str, obj.get(self.aliases.type))
            lang = obj.get(self.aliases.lang)
        else:
            kdef: Optional[StrOrObject] = None
            if via_key is not None and via_key in self.context:
                kdef = cast(StrOrObject, self.context[via_key])
            coerce_to: Optional[str] = None
            if isinstance(kdef, Dict) and TYPE in kdef:
                coerce_to = cast(str, kdef[TYPE])
            if coerce_to == VOCAB:
                next = False
                for v in as_list(value):
                    if next:
                        write(' , ')
                    else:
                        next = True
                    write(self.ref_repr(v, True) if isinstance(v, str) else self.to_str(v))
                return
            elif coerce_to == ID:
                next = False
                for v in as_list(value):
                    if next:
                        write(' , ')
                    else:
                        next = True
                    write(self.ref_repr(v))
                return
            elif coerce_to:
                datatype = coerce_to
            else:
                if isinstance(kdef, Dict) and LANGUAGE in kdef:
                    lang = kdef[LANGUAGE]

        next = False
        for v in as_list(value):
            if next:
                write(' , ')
            else:
                next = True

            write(self.to_str(v, datatype, lang))

    def to_str(self, v: object, datatype: Optional[str] = None, lang: Optional[object] = None) -> str:
        if isinstance(v, str):
            parts: List[str] = []
            escaped = v.replace('\\', '\\\\')
            quote = '"'
            if escaped.find('\n') > -1:
                quote = '"""'
                if escaped.endswith('"'):
                    escaped = f'{escaped[0 : len(escaped) - 1]}\\"'
            else:
                escaped = escaped.replace('"', '\\"')
            parts.append(quote)
            parts.append(escaped)
            parts.append(quote)
            if datatype:
                parts.append("^^" + self.to_valid_term(cast(str, self.term_for(datatype))))
            elif isinstance(lang, str):
                parts.append(f"@{lang}")
            return ''.join(parts)
        elif isinstance(v, bool):
            return 'true' if v else 'false'
        else: # int or float
            return str(v)

    def term_for(self, key: str) -> Optional[str]:
        if key.startswith("@"):
            return key
        elif key.find(":") > -1 or \
                             key.find('/') > -1 or \
                             key.find('#') > -1:
            return key
        elif key in self.context:
            kdef = self.context[key]
            if kdef is None:
                return None
            term = None
            if isinstance(kdef, Dict):
                term = kdef.get(ID, key)
            else:
                term = kdef
            assert isinstance(term, str)
            ci: int = term.find(":")
            return f":{term}" if ci == -1 else term
        else:
            return ":" + key

    def rev_key_for(self, key: str) -> Optional[str]:
        kdef = self.context[key]
        if isinstance(kdef, Dict) and REVERSE in kdef:
            return cast(str, kdef[REVERSE])
        return None

    def index_key_for(self, key: str) -> Optional[str]:
        kdef: object = self.context.get(key)
        if isinstance(kdef, Dict) and kdef.get(CONTAINER) == INDEX:
            return cast(str, kdef.get(ID, key))
        return None

    def make_top_object(self, s: Optional[str], rev_key: str, it: Dict) -> StrObject:
        node = dict(it)
        # TODO: probe object to find an id:d top object...
        if self.aliases.id not in node:
            node[self.aliases.id] = f'_:bnode-{self.bnode_counter}'
            self.bnode_counter += 1
        node[rev_key] = { self.aliases.id: s }
        return node

    def repr_type(self, t: StrOrObject) -> str:
        tstr: str = t if isinstance(t, str) else cast(str, cast(Dict, t)[TYPE]) # assuming annotation form
        return self.to_valid_term(cast(str, self.term_for(tstr)))

    def ref_repr(self, refobj: Optional[StrOrObject], use_vocab = False) -> str:
        if refobj is None:
            return '[]'

        if isinstance(refobj, Dict) and self.aliases.id in refobj:
            return self.repr_triple(refobj)

        ref = cast(str, refobj)

        c_i: int = ref.find(':')
        if c_i > -1:
            pfx = ref[0 : c_i]
            if pfx == "_":
                node_id: str = ref + self.unique_bnode_suffix
                if self.bnode_skolem_base:
                    ref = self.bnode_skolem_base + node_id[2 :]
                else:
                    return self.to_valid_term(node_id)
            elif pfx in self.context:
                local = ref[c_i + 1 :]
                return f'{pfx}:{self.escape_pname_local(local)}'
        elif use_vocab and ref.find("/") == -1:
            return ":" + ref

        if VOCAB in self.context and ref.startswith(cast(str, self.context[VOCAB])):
            return ":" + ref[len(cast(str, self.context[VOCAB])) :]

        ref = self.clean_value(ref)

        c_i = ref.find(':')
        if c_i > -1:
            pfx = ref[0 : c_i]
            rest = ref[c_i :]
            if pfx in self.context:
                return ref
            # non-std: check if ref is "most likely" a pname
            if len(self.context) > 0 and \
                    rest.find(':') == -1 and \
                    WORD_START.match(rest) is not None and \
                    WORD_START.match(pfx) is not None:
                return ref

        return f'<{ref}>'

    def repr_triple(self, ref: StrObject) -> str:
        if self.settings.drop_rdfstar:
            raise Exception('Triple nodes disallowed unless in RDF-star mode')

        s: str = self.ref_repr(cast(str, ref[self.aliases.id]))

        p: str = ''
        obj: StrOrObject = ''
        for k in ref.keys():
            if k == self.aliases.id:
                continue
            if p != '':
                raise Exception('Quoted triples cannot contain multiple statements')

            p = cast(str, self.term_for(k))
            obj = cast(StrObject, ref[k])

        o: str
        if p == self.aliases.type:
            p = "a"
            o = self.repr_type(obj)
        else:
            if isinstance(obj, List):
                raise Exception('Quoted triples must have one single object')
            if self.is_lang_container(p) and isinstance(obj, Dict):
                raise Exception('Language containers not yet supported in quoted triples')
            if isinstance(obj, Dict) and self.aliases.list_ in obj:
                raise Exception('Quoted triples cannot contain Lists')

            if not isinstance(obj, Dict) or self.aliases.value in obj:
                l: List[str] = []
                self.to_literal(obj, p, lambda x: l.append(cast(str, x)))
                o = ''.join(l)
            else:
                assert isinstance(obj, Dict) and self.aliases.id in obj
                o = self.ref_repr(cast(str, obj[self.aliases.id]))

        return f'<< {s} {p} {o} >>'

    def to_valid_term(self, term: str) -> str:
        term = self.clean_value(term)
        c_i: int = term.find(':')
        pfx: Optional[str] = term[0 : c_i] if c_i > -1 else None
        if (
                not (pfx in self.context) and
                # non-std: fake pnames even when missing prefixes!
                (term.find('/') > -1 or
                 term.find('#') > -1 or
                 pfx is not None and term.rfind(':') > len(pfx))
        ):
            return f'<{term}>'
        if pfx is not None:
            local = term[c_i + 1 :]
            return f'{pfx}:{self.escape_pname_local(local)}'
        return self.escape_pname_local(term)

    def has_keys(self, obj: StrObject, at_least: int = 1) -> bool:
        seen = 0
        for k in obj.keys():
            if k != self.aliases.annotation:
                seen += 1
                if seen == at_least:
                    return True
        return False

    def clean_value(self, v: str) -> str:
        return v

    def escape_pname_local(self, pnlocal: str) -> str:
        # From: https://www.w3.org/TR/turtle/#grammar-production-PN_LOCAL_ESC
        # Note that '.-' are OK within, but need escaping at start/end.
        # And '_' *may* be escaped but is OK everywhere in PN_LOCAL.
        return PNAME_LOCAL_ESC.sub(r'\\\1', pnlocal)

    def gen_skolem_id(self) -> Optional[str]:
        if self.bnode_skolem_base is None:
            return None
        return self.bnode_skolem_base + uuid4()

    def get_indent(self, depth: int) -> str:
        chunks: List[str] = []
        i = -1
        while i < depth:
            i += 1
            chunks.append(self.settings.indent_chars)
        return ''.join(chunks)

    def write(self, s: str):
        self.out.write(s if s is not None else '')

    def writeln(self, s: Optional[str] = None):
        self.out.write((s if s is not None else '') + '\n')

    # TODO: for transpile (refactor to avoid passing a write function above)
    def _write(self, s) -> Optional[str]:
        self.write(cast(str, s))
        return None


def collect_prefixes(context: Optional[object]) -> Dict[str, str]:
    if not isinstance(context, Dict):
        return {}

    prefixes = {}
    for key, value in cast(StrObject, context).items():
        if isinstance(value, str) and value[-1] in PREFIX_DELIMS:
            prefixes['' if key == VOCAB else key] = value
        elif isinstance(value, Dict) and value.get(PREFIX) == True:
            prefixes[key] = value[ID]

    return prefixes


def as_list(value) -> List:
    return value if isinstance(value, List) else [value]


if __name__ == '__main__':
    import sys, json

    data = json.load(sys.stdin)
    out = Output(sys.stdout)
    if '--turtle' in sys.argv:
        serialize_turtle(data, out, union='--union' in sys.argv)
    else:
        serialize(data, out)



from typing import Optional, Tuple, Dict, List, Union, cast
from ..base import ID, GRAPH, REVERSE, JsonMap, as_list


def make_index(graph: List[JsonMap], add_reverses=True) -> Dict[str, JsonMap]:
    index: Dict[str, JsonMap] = {}

    for item in graph:
        id: Optional[str] = cast(Optional[str], item.get(ID))
        if isinstance(id, str):
            index[id] = item

    if add_reverses:
        _index_reverses(index)

    return index


def _index_reverses(index: Dict[str, JsonMap]):
    for item in index.values():
        if ID not in item:
            continue
        for link in item.keys():
            refs: List = as_list(item[link])
            for ref in refs:
                if not isinstance(ref, Dict):
                    continue
                linked: object = index.get(cast(str, ref.get(ID)))
                if not isinstance(linked, Dict):
                    continue
                revmap: Dict = linked.setdefault(REVERSE, {})
                revs: List[Dict] = revmap.setdefault(link, [])
                if not any(rev[ID] == item[ID] for rev in revs):
                    revs.append({ID: item[ID]})



from .base import *
from .context import Context, Term


def get_inverse_context(active_context: Context) -> Dict:
    if active_context._inverse_context is None:
        active_context._inverse_context = create_inverse_context(active_context)
    return active_context._inverse_context


def create_inverse_context(active_context: Context) -> JsonMap:
    # 1)
    result: JsonMap = {}

    # 2)
    default_language: str = active_context.default_language if active_context.default_language else NONE

    # 3)
    term_keys: List[Optional[str]] = list(active_context.terms.keys())
    term_keys.sort(key=lambda s: (len(cast(str, s)) if s is not None else 0, s))
    for term_key in term_keys:
        assert isinstance(term_key, str) # TODO: Optional just to allow .get(Optional) ...
        term_dfn: Optional[Term] = active_context.terms[term_key]
        # 3.1)
        if term_dfn is None:
            continue
        # 3.2)
        container: str = NONE
        if len(term_dfn.container) > 0:
            container = ''.join(sorted(term_dfn.container))
        # 3.3)
        iri: str = term_dfn.iri
        # 3.4)
        if iri not in result:
            result[iri] = {}
        # 3.5)
        container_map: JsonMap = cast(JsonMap, result[iri])
        # 3.6)
        if container not in container_map:
            container_map[container] = {LANGUAGE: {}, TYPE: {}, ANY: {NONE: term_key}}
        # 3.7)
        typelanguage_map: JsonMap = cast(JsonMap, container_map[container])
        # 3.8)
        type_map: JsonMap = cast(JsonMap, typelanguage_map[TYPE])
        # 3.9)
        language_map: JsonMap = cast(JsonMap, typelanguage_map[LANGUAGE])
        lang_dir: str
        # 3.10)
        if term_dfn.is_reverse_property:
            # 3.10.1)
            if REVERSE not in type_map:
                type_map[REVERSE] = term_key
        # 3.11)
        elif term_dfn.type_mapping == NONE:
            # 3.11.1)
            if ANY not in language_map:
                language_map[ANY] = term_key
            # 3.11.2)
            if ANY not in type_map:
                type_map[ANY] = term_key
        # 3.12)
        elif term_dfn.type_mapping is not None:
            # 3.12.1)
            if term_dfn.type_mapping not in type_map:
                type_map[term_dfn.type_mapping] = term_key
        # 3.13)
        elif term_dfn.language is not None and term_dfn.direction is not None:
            # 3.13.1 + # 3.13.5)
            lang_dir = NULL
            # 3.13.2)
            if term_dfn.language is not NULL and term_dfn.direction is not NULL:
                lang_dir = f'{term_dfn.language}_{term_dfn.direction}'
            # 3.13.3)
            elif term_dfn.language is not NULL:
                lang_dir = term_dfn.language
            # 3.13.4)
            elif term_dfn.direction is not NULL:
                lang_dir = term_dfn.direction
            # 3.13.6)
            if lang_dir not in language_map:
                language_map[lang_dir] = term_key
        # 3.14)
        elif term_dfn.language is not None:
            # 3.14.1)
            language: str = term_dfn.language
            # 3.14.2)
            if language not in language_map:
                language_map[language] = term_key
        # 3.15)
        elif term_dfn.direction is not None:
            # 3.15.1)
            direction: str = term_dfn.direction
            if direction == NULL:
                direction = NONE
            else:
                direction = f'_{direction}'
            # 3.15.2)
            if direction not in language_map:
                language_map[direction] = term_key
        # 3.16)
        elif active_context.default_base_direction:
            # 3.16.1)
            lang_dir = f'{active_context.default_language}_{active_context.default_base_direction}'
            # 3.16.2)
            if lang_dir not in language_map:
                language_map[lang_dir] = term_key
            # 3.16.3)
            if NONE not in language_map:
                language_map[NONE] = term_key
            # 3.16.4)
            if NONE not in type_map:
                type_map[NONE] = term_key
        # 3.17)
        else:
            # 3.17.1)
            if active_context.default_language  is not None and active_context.default_language not in language_map:
                language_map[active_context.default_language] = term_key
            # 3.17.2)
            if NONE not in language_map:
                language_map[NONE] = term_key
            # 3.17.3)
            if NONE not in type_map:
                type_map[NONE] = term_key

    # 4)
    return result




from typing import Iterator, List, Dict, Union, NamedTuple, cast
from pathlib import Path
from ..common import Input, dump_canonical_json
from ..jsonld.base import CONTEXT, GRAPH, ID, TYPE, LIST
from ..jsonld.expansion import expand
from ..jsonld.compaction import compact
from ..jsonld.flattening import flatten
from ..jsonld.rdf import to_rdf_dataset, to_jsonld
from ..nq import parser as nq_parser

from . import parser

##
# Official test suite located at: https://www.w3.org/2013/TrigTests/TESTS.tar.gz


class TestCase(NamedTuple):
    ttype: str
    taction: str
    tresult: str


def read_manifest(manifest_path) -> Iterator[TestCase]:
    data = cast(Dict, parser.parse(Input(str(manifest_path))))
    index = {node[ID]: node for node in data[GRAPH] if ID in node}

    testentries = cast(List[Dict], index['']['mf:entries'][LIST])
    for testentry in testentries:
        tnode = index[testentry[ID]]
        ttype = tnode['rdf:type'][ID] if 'rdf:type' in tnode else tnode[TYPE]
        taction = tnode['mf:action'][ID]
        tresult = tnode['mf:result'][ID] if 'mf:result' in tnode else None
        yield TestCase(ttype, taction, tresult)


def run_tests(test_suite_dir: Union[str, Path]):
    test_suite_dir = Path(test_suite_dir)

    i = 0
    failed = 0
    passed = 0

    manifest = read_manifest(test_suite_dir / 'manifest.ttl')

    for ttype, taction, tresult in manifest:
        i += 1
        trig_path = test_suite_dir / taction
        negative = ttype == 'rdft:TestTrigNegativeSyntax'
        inp = Input(str(trig_path))
        try:
            result = cast(Dict, parser.parse(inp))
            assert result is not None
        except Exception as e:
            if negative:
                passed += 1
            else:
                print(f'FAILED on {trig_path} (a {ttype}):', e)
                failed += 1
        else:
            if negative:
                print(f'SHOULD FAIL on {trig_path} (a {ttype})')
                failed += 1
            elif tresult:
                nq_path = test_suite_dir / tresult
                try:
                    expected = nq_parser.parse(Input(str(nq_path)))
                except Exception:
                    print(f'Error parsing NQuads {nq_path}')
                    raise

                base_uri = 'http://www.w3.org/2013/TriGTests/'
                context = result[CONTEXT]
                resultrepr = datarepr(result, context, f'{base_uri}{trig_path.name}')
                expectedrepr = datarepr(expected, context, f'{base_uri}{nq_path.name}')

                if resultrepr != expectedrepr:
                    print(f'FAILED COMPARISON for {trig_path} ({ttype}). Got:\n',
                          f'{resultrepr}\n',
                          f'Expected from {nq_path}:\n', f'{expectedrepr}', sep='\t')
                    print()
                    failed += 1
                else:
                    passed += 1
            else:
                passed += 1

    print(f'Ran {i} tests. Passed {passed}, failed {failed}')


def datarepr(data, context, base_uri) -> str:
    data = expand(data, base_uri)
    data = flatten(data)
    dataset = to_rdf_dataset(data)
    data = to_jsonld(dataset)
    data = compact(context, data, '', ordered=True)
    return dump_canonical_json(data)


if __name__ == '__main__':
    import sys
    args = sys.argv[1:]

    test_suite_dir = args.pop(0) if args else 'cache/trig-tests'

    run_tests(test_suite_dir)
